from __future__ import annotations

from briefcase.commands import (
    BuildCommand,
    CreateCommand,
    OpenCommand,
    PackageCommand,
    PublishCommand,
    RunCommand,
    UpdateCommand,
)
from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseConfigError
from briefcase.integrations.flatpak import Flatpak
from briefcase.platforms.linux import LinuxMixin


class LinuxFlatpakMixin(LinuxMixin):
    output_format = "flatpak"
    supported_host_os = {"Linux"}
    supported_host_os_reason = "Flatpaks can only be built on Linux."

    def binary_path(self, app):
        # Flatpak doesn't really produce an identifiable "binary" as part of its
        # build process, so the SDK wrapper creates a file that can use to identify
        # if run has been invoked. As a neat side effect, it's also a shell script
        # that can invoke the flatpak.
        return self.bundle_path(app) / app.bundle_identifier

    def project_path(self, app):
        return self.bundle_path(app)

    def distribution_path(self, app):
        binary_name = app.formal_name.replace(" ", "_")
        return (
            self.dist_path
            / f"{binary_name}-{app.version}-{self.tools.host_arch}.flatpak"
        )

    def verify_tools(self):
        """Verify that we're on Linux."""
        super().verify_tools()
        Flatpak.verify(tools=self.tools)

    def flatpak_runtime_repo(self, app):
        try:
            repo_url = app.flatpak_runtime_repo_url
            try:
                repo_alias = app.flatpak_runtime_repo_alias
            except AttributeError:
                raise BriefcaseConfigError(
                    "If you specify a custom Flatpak runtime repository, "
                    "you must also specify an alias for that repository using "
                    "`flatpak_runtime_repo_alias`"
                )

        except AttributeError:
            repo_alias = Flatpak.DEFAULT_REPO_ALIAS
            repo_url = Flatpak.DEFAULT_REPO_URL

        return repo_alias, repo_url

    def flatpak_runtime(self, app):
        try:
            return app.flatpak_runtime
        except AttributeError as e:
            raise BriefcaseConfigError(
                """\
The App does not specify the Flatpak runtime to use.

Your application configuration must provide values for
`flatpak_sdk`, `flatpak_runtime`, and `flatpak_runtime_version`.
"""
            ) from e

    def flatpak_runtime_version(self, app):
        try:
            return app.flatpak_runtime_version
        except AttributeError as e:
            raise BriefcaseConfigError(
                """\
The App does not specify the version of the Flatpak runtime to use.

Your application configuration must provide values for
`flatpak_sdk`, `flatpak_runtime`, and `flatpak_runtime_version`.
"""
            ) from e

    def flatpak_sdk(self, app):
        try:
            return app.flatpak_sdk
        except AttributeError as e:
            raise BriefcaseConfigError(
                """\
The App does not specify the Flatpak SDK to use.

Your application configuration must provide values for
`flatpak_sdk`, `flatpak_runtime`, and `flatpak_runtime_version`.
"""
            ) from e


class LinuxFlatpakCreateCommand(LinuxFlatpakMixin, CreateCommand):
    description = "Create and populate a Linux Flatpak."
    hidden_app_properties = {"permission", "finish_arg"}

    def output_format_template_context(self, app: AppConfig):
        """Add flatpak runtime/SDK details to the app template."""
        return {
            "flatpak_runtime": self.flatpak_runtime(app),
            "flatpak_runtime_version": self.flatpak_runtime_version(app),
            "flatpak_sdk": self.flatpak_sdk(app),
        }

    def permissions_context(self, app: AppConfig, x_permissions: dict[str, str]):
        """Additional template context for permissions.

        :param app: The config object for the app
        :param x_permissions: The dictionary of known cross-platform permission
            definitions.
        :returns: The template context describing permissions for the app.
        """
        # The default finish arguments that Briefcase adds on every Flatpak.
        finish_args = {
            # X11 + XShm access
            "share=ipc": True,
            "socket=x11": True,
            # Disable Wayland access
            "nosocket=wayland": True,
            # Network access
            "share=network": True,
            # GPU access
            "device=dri": True,
            # Sound access
            "socket=pulseaudio": True,
            # Host filesystem access
            "filesystem=xdg-cache": True,
            "filesystem=xdg-config": True,
            "filesystem=xdg-data": True,
            "filesystem=xdg-documents": True,
            # DBus access
            "socket=session-bus": True,
        }

        finish_args.update(getattr(app, "finish_arg", {}))

        return {
            "finish_args": finish_args,
        }


class LinuxFlatpakUpdateCommand(LinuxFlatpakCreateCommand, UpdateCommand):
    description = "Update an existing Linux Flatpak."


class LinuxFlatpakOpenCommand(LinuxFlatpakMixin, OpenCommand):
    description = "Open the folder containing an existing Linux Flatpak project."


class LinuxFlatpakBuildCommand(LinuxFlatpakMixin, BuildCommand):
    description = "Build a Linux Flatpak."

    def build_app(self, app: AppConfig, **kwargs):
        """Build an application.

        :param app: The application to build
        """
        self.logger.info(
            "Ensuring Flatpak runtime for the app is available...",
            prefix=app.app_name,
        )
        flatpak_repo_alias, flatpak_repo_url = self.flatpak_runtime_repo(app)

        with self.input.wait_bar("Ensuring Flatpak runtime repo is registered..."):
            self.tools.flatpak.verify_repo(
                repo_alias=flatpak_repo_alias,
                url=flatpak_repo_url,
            )

        # ``flatpak install`` uses a lot of console animations, and there
        # doesn't appear to be a way to turn off these animations. Use those
        # native animations rather than wrapping in a wait_bar.
        self.tools.flatpak.verify_runtime(
            repo_alias=flatpak_repo_alias,
            runtime=self.flatpak_runtime(app),
            runtime_version=self.flatpak_runtime_version(app),
            sdk=self.flatpak_sdk(app),
        )

        self.logger.info("Building Flatpak...", prefix=app.app_name)
        with self.input.wait_bar("Building..."):
            self.tools.flatpak.build(
                bundle_identifier=app.bundle_identifier,
                app_name=app.app_name,
                path=self.bundle_path(app),
            )


class LinuxFlatpakRunCommand(LinuxFlatpakMixin, RunCommand):
    description = "Run a Linux Flatpak."

    def run_app(
        self,
        app: AppConfig,
        test_mode: bool,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param test_mode: Boolean; Is the app running in test mode?
        :param passthrough: The list of arguments to pass to the app
        """
        # Set up the log stream
        kwargs = self._prepare_app_kwargs(app=app, test_mode=test_mode)

        # Starting a flatpak has slightly different startup arguments; however,
        # the rest of the app startup process is the same. Transform the output
        # of the "default" behavior to be in flatpak format.
        if test_mode:
            kwargs = {"main_module": kwargs["env"]["BRIEFCASE_MAIN_MODULE"]}
        else:
            kwargs = {}

        # Console apps must operate in non-streaming mode so that console input can
        # be handled correctly. However, if we're in test mode, we *must* stream so
        # that we can see the test exit sentinel
        if app.console_app and not test_mode:
            self.logger.info("=" * 75)
            self.tools.flatpak.run(
                bundle_identifier=app.bundle_identifier,
                args=passthrough,
                stream_output=False,
                **kwargs,
            )
        else:
            # Start the app in a way that lets us stream the logs
            app_popen = self.tools.flatpak.run(
                bundle_identifier=app.bundle_identifier,
                args=passthrough,
                stream_output=True,
                **kwargs,
            )

            # Start streaming logs for the app.
            self._stream_app_logs(
                app,
                popen=app_popen,
                test_mode=test_mode,
                clean_output=False,
            )


class LinuxFlatpakPackageCommand(LinuxFlatpakMixin, PackageCommand):
    description = "Package a Linux Flatpak for distribution."

    def package_app(self, app: AppConfig, **kwargs):
        """Start the application.

        :param app: The config object for the app
        """
        self.logger.info("Building bundle...", prefix=app.app_name)
        with self.input.wait_bar("Bundling..."):
            _, flatpak_repo_url = self.flatpak_runtime_repo(app)
            self.tools.flatpak.bundle(
                repo_url=flatpak_repo_url,
                bundle_identifier=app.bundle_identifier,
                app_name=app.app_name,
                version=app.version,
                build_path=self.bundle_path(app),
                output_path=self.distribution_path(app),
            )


class LinuxFlatpakPublishCommand(LinuxFlatpakMixin, PublishCommand):
    description = "Publish a Linux Flatpak."


# Declare the briefcase command bindings
create = LinuxFlatpakCreateCommand
update = LinuxFlatpakUpdateCommand
open = LinuxFlatpakOpenCommand
build = LinuxFlatpakBuildCommand
run = LinuxFlatpakRunCommand
package = LinuxFlatpakPackageCommand
publish = LinuxFlatpakPublishCommand



from subprocess import CalledProcessError
from unittest import mock

from briefcase.integrations.xcode import XcodeCliTools


def test_verify_command_line_tools_install(mock_tools):
    """Xcode CLI tools can be verified."""
    mock_tools.subprocess.check_output.side_effect = [
        CalledProcessError(cmd=["xcode-select", "--install"], returncode=1),
        "clang 37.42",  # clang --version
    ]

    XcodeCliTools.verify(mock_tools)

    # The command line tools are verified
    assert mock_tools.xcode_cli is not None


def test_reverify_command_line_tools_install(mock_tools):
    """A second call to verify is a no-op."""

    xcode_cli = mock.MagicMock()
    mock_tools.xcode_cli = xcode_cli

    XcodeCliTools.verify(mock_tools)

    # The command line tools are verified
    assert mock_tools.xcode_cli is not None



import subprocess

import pytest

from briefcase.integrations.windows_sdk import WindowsSDK

from .test_WindowsSDK__verify import setup_winsdk_install


@pytest.fixture
def windows_sdk(mock_tools, tmp_path, monkeypatch) -> WindowsSDK:
    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "83.0")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    yield WindowsSDK(
        tools=mock_tools,
        root_path=tmp_path / "win_sdk",
        version="83.1.0.0",
        arch="x64",
    )


def test_winsdk_signtool_succeeds(windows_sdk, tmp_path):
    """Validation succeeds if signtool exists and can be run."""
    setup_winsdk_install(
        base_path=tmp_path,
        version=".".join(windows_sdk.version.split(".")[0:3]),
        arch="x64",
    )

    assert WindowsSDK._verify_signtool(windows_sdk) is True


def test_winsdk_signtool_does_not_exist(windows_sdk):
    """Validation fails if signtool does not exist."""
    assert WindowsSDK._verify_signtool(windows_sdk) is False


def test_winsdk_signtool_raises_oserror(windows_sdk, tmp_path):
    """Validation fails if running signtool raises a OSError."""
    setup_winsdk_install(
        base_path=tmp_path,
        version=".".join(windows_sdk.version.split(".")[0:3]),
        arch="x64",
    )
    windows_sdk.tools.subprocess.check_output.side_effect = OSError(
        14001,
        " The application has failed to start because its side-by-side configuration is incorrect.",
    )

    assert WindowsSDK._verify_signtool(windows_sdk) is False


def test_winsdk_signtool_raises_calledprocesserror(windows_sdk, tmp_path):
    """Validation fails if running signtool raises a OSError."""
    setup_winsdk_install(
        base_path=tmp_path,
        version=".".join(windows_sdk.version.split(".")[0:3]),
        arch="x64",
    )
    windows_sdk.tools.subprocess.check_output.side_effect = (
        subprocess.CalledProcessError(
            returncode=1,
            cmd="signtool.exe -?",
            output="Unknown error",
        )
    )

    assert WindowsSDK._verify_signtool(windows_sdk) is False



from __future__ import annotations

import re
import sys
from functools import cached_property, partial
from pathlib import Path
from shutil import copy2, copytree
from tempfile import TemporaryDirectory
from urllib.parse import urlparse

from packaging.utils import canonicalize_name

from ..config import is_valid_app_name
from .new import NewCommand, parse_project_overrides, titlecase

if sys.version_info >= (3, 11):  # pragma: no-cover-if-lt-py311
    import tomllib
else:  # pragma: no-cover-if-gte-py311
    import tomli as tomllib

from briefcase.config import make_class_name
from briefcase.exceptions import BriefcaseCommandError


class ConvertCommand(NewCommand):
    cmd_line = "briefcase convert"
    command = "convert"
    platform = "all"
    output_format = None
    description = "Set up an existing project for Briefcase."

    @cached_property
    def pyproject(self):
        """The contents of the pyproject.toml file (as a dictionary)."""
        pyproject_file = self.base_path / "pyproject.toml"
        if pyproject_file.exists():
            with open(pyproject_file, "rb") as file:
                return tomllib.load(file)
        else:
            return {}

    @cached_property
    def pep621_data(self):
        """The pyproject["project"] dictionary.

        Empty dict if [project] is not in the pyproject.toml file.
        """
        return self.pyproject.get("project", {})

    def validate_test_source_dir(self, module_name: str, test_source_dir: str) -> bool:
        """Determine if the test_source_dir is valid.

        :param test_source_dir: The candidate test source directory
        :returns: True. If there are any validation problems, raises ValueError with a
            diagnostic message.
        """
        test_path = self.base_path / test_source_dir
        if (test_entry := test_path / f"{module_name}.py").exists():
            raise ValueError(
                f"'{test_entry}' is reserved for the briefcase test entry script, but it already exists.\n"
                "\n"
                "Briefcase expects this file to contain the test entry script, so if "
                f"'{self.base_path / test_source_dir}' is your test directory, you must "
                f"rename '{test_entry}' before setting up your project for briefcase."
            )

        return True

    def validate_source_dir(self, module_name: str, source_dir: str) -> bool:
        """Determine if the source_dir is valid.

        :param source_dir: The candidate source directory
        :returns: True. If there are any validation problems, raises ValueError with a
            diagnostic message.
        """
        if (self.base_path / source_dir).name != module_name:
            raise ValueError(
                f"The source directory '{self.base_path / source_dir}' must have "
                f"the same name as the app, but with underscores instead of dashes "
                f"('{module_name}'), not '{(self.base_path / source_dir).name}'."
            )

        if not (self.base_path / source_dir).is_dir():
            raise ValueError(
                "The source directory must exist and contain a '__main__.py' file."
            )

        if not (self.base_path / source_dir / "__main__.py").is_file():
            raise ValueError("The source directory must contain a '__main__.py' file")

        return True

    def _validate_existing_app_name(self, candidate):
        """Override default validation to allow app names with a pre-existing
        directory."""
        pass

    def input_app_name(self, override_value: str | None) -> str:
        """Ask about the app name, using hints from the pyproject.toml file or directory
        name if there are any.

        If the name is specified in a PEP621 compliant pyproject.toml file, then it is
        used without prompting.

        :returns: The app name
        """
        intro = (
            "We need a name that can serve as a machine-readable Python package name for "
            "your application. This name must be PEP508-compliant - that means the name "
            "may only contain letters, numbers, hyphens and underscores; it can't contain "
            "spaces or punctuation, and it can't start with a hyphen or underscore."
        )

        default = "hello-world"
        if (
            "name" in self.pep621_data
            and is_valid_app_name(self.pep621_data["name"])
            and override_value is None
        ):
            app_name = canonicalize_name(self.pep621_data["name"])
            self.prompt_divider(title="App name")
            self.input.prompt()
            self.input.prompt(
                f"Using value from PEP621 formatted pyproject.toml {app_name!r}"
            )
            return app_name

        if is_valid_app_name(self.base_path.name):  # Directory name is normalised
            default = canonicalize_name(self.base_path.name)
            intro += (
                "\n\n"
                f"Based on your PEP508 formatted directory name, we suggest an "
                f"app name of '{default}', but you can use another name if you want."
            )

        return self.input_text(
            intro=intro,
            variable="app name",
            default=default,
            validator=self.validate_app_name,
            override_value=override_value,
        )

    def input_formal_name(self, app_name: str, override_value: str | None) -> str:
        """Ask about formal name.

        :returns: The source directory
        """
        default = titlecase(" ".join(re.split("[-_]", app_name)))
        return self.input_text(
            intro=(
                "We need a formal name for your application. This is the name that "
                "will be displayed to humans whenever the name of the application is "
                "displayed. It can include spaces and punctuation, and any "
                "capitalization will be used as you type it here.\n"
                "\n"
                f"Based on the app name, we suggest a formal name of '{default}', but "
                "you can use another name if you want."
            ),
            variable="formal name",
            default=default,
            override_value=override_value,
        )

    def get_source_dir_hint(self, app_name: str, module_name: str) -> tuple[str, str]:
        """Parse folder layout to get hint for the source directory.

        :returns: The source directory hint
        :returns: The description text for the source dir prompt.
        """
        possible_src_dirs = [
            p.parent for p in self.base_path.glob(f"**/{module_name}/__main__.py")
        ]
        if (self.base_path / f"src/{module_name}") in possible_src_dirs:
            default = self.base_path / f"src/{module_name}"
        elif (self.base_path / module_name) in possible_src_dirs:
            default = self.base_path / module_name
        elif possible_src_dirs:
            default = min(
                possible_src_dirs,
                key=lambda p: (len(p.parents), len(str(p))),
            )
        else:  # We have already checked that there are directories in the project root
            raise BriefcaseCommandError(
                "Cannot find a suitable source directory for the app.\n"
                "\n"
                f"Based on your app name, '{app_name}', you must have a directory "
                f"named '{module_name}', either in your project root or a subdirectory "
                f"(e.g. 'src/{module_name}'), that contains a '__main__.py'."
            )

        default = str(default.relative_to(self.base_path)).replace("\\", "/")
        intro = (
            "To set up an existing project for Briefcase, we need to know the path "
            "of the application entry point relative to the project root (the current "
            "working directory).\n"
            "\n"
            "For example, if you have an existing project 'myapp', and you can start "
            "'myapp' by running 'src/myapp/__main__.py', then you should set the source "
            "directory to 'src/myapp'.\n"
            "\n"
            f"Based on your project's folder layout, we believe it might be '{default}'."
        )
        return default, intro

    def input_source_dir(
        self,
        app_name: str,
        module_name: str,
        override_value: str | None,
    ) -> str:
        """Ask about the source dir, using hints from the project folder layout.

        :returns: The source directory
        """
        default, intro = self.get_source_dir_hint(app_name, module_name)
        return self.input_text(
            intro=intro,
            variable="source directory",
            default=default,
            validator=partial(self.validate_source_dir, module_name),
            override_value=override_value,
        )

    def input_test_source_dir(self, module_name, override_value) -> str:
        """Ask about the test source dir, using hints from the project folder layout.

        :returns: The test source directory
        """
        intro = (
            "We also need to know the path to the test suite (if it exists). The test path "
            "should be relative to the project root directory.\n"
            "\n"
            "If the provided directory doesn't exist, it will be created and populated with "
            "some default test files."
        )
        if (self.base_path / "tests").is_dir():
            default = "tests"
            intro += (
                "\n\nBased on your project's folder structure, we believe "
                + "'tests' might be your test directory"
            )
        elif (self.base_path / "test").is_dir():
            default = "test"
            intro += (
                "\n\nBased on your project's folder structure, we believe "
                "'test' might be your test directory"
            )
        else:
            default = "tests"

        return self.input_text(
            intro=intro,
            variable="test source directory",
            default=default,
            validator=partial(self.validate_test_source_dir, module_name),
            override_value=override_value,
        )

    def input_description(self, override_value: str | None) -> str:
        """Ask about the app description, using hints from the pyproject.toml file if
        there are any.

        If the description is specified in a PEP621 compliant pyproject.toml file, then
        it is used without prompting.

        :returns: The app description
        """
        if "description" in self.pep621_data and override_value is None:
            description = self.pep621_data["description"]

            self.prompt_divider(title="Description")
            self.input.prompt()
            self.input.prompt(
                f"Using value from PEP621 formatted pyproject.toml {description!r}"
            )
            return description

        return self.input_text(
            intro="Now, we need a one line description for your application.",
            variable="description",
            default="My first application",
            override_value=override_value,
        )

    def input_url(self, app_name, override_value: str | None) -> str:
        """Ask about the URL, using hints from the pyproject.toml file if there are any.

        :returns: The project
        """
        options = list(self.pep621_data.get("urls", {}).values())

        if not options or override_value:
            default = self.make_project_url("com.example", app_name)
            return self.input_text(
                intro=(
                    "What website URL do you want to use for this application? Based "
                    f"on your existing 'pyproject.toml', this might be {default}"
                ),
                variable="application URL",
                default=default,
                validator=self.validate_url,
                override_value=override_value,
            )

        options.append("Other")
        url = self.select_option(
            intro=(
                "What website URL do you want to use for this application? The "
                "following URLs are defined in your existing 'pyproject.toml'; "
                "select 'Other' to provide a different URL."
            ),
            variable="application URL",
            default=None,
            options=options,
            override_value=override_value,
        )

        if url == "Other":
            url = self.input_text(
                intro="What website URL do you want to use for the application?",
                variable="application URL",
                default=self.make_project_url("com.example", app_name),
                validator=self.validate_url,
            )

        return url

    def input_bundle(self, url, app_name, override_value: str | None) -> str:
        default = ".".join(reversed(urlparse(url).netloc.split(".")))
        return self.input_text(
            intro=(
                "Now we need a bundle identifier for your application.\n"
                "\n"
                "App stores need to protect against having multiple applications with "
                "the same name; the bundle identifier is the namespace they use to "
                "identify applications that come from you. The bundle identifier is "
                "usually the domain name of your company or project, in reverse "
                "order.\n"
                "\n"
                "Based on the application URL you selected, it looks like your bundle "
                f"should be '{default}'. The bundle will be combined with your "
                "application's machine readable name to form a complete application "
                f"identifier ('{default}.{app_name}')."
            ),
            variable="bundle identifier",
            default=default,
            validator=self.validate_bundle,
            override_value=override_value,
        )

    def input_author(self, override_value: str | None) -> str:
        """Ask about the author name, using hints from the pyproject.toml file if there
        are any.

        :returns: author name
        """
        intro = (
            "Who do you want to be credited as the author of this application? "
            "This could be your own name, or the name of your company you work for."
        )

        options = [
            author["name"]
            for author in self.pep621_data.get("authors", [])
            if "name" in author
        ]

        if not options or override_value is not None:
            return self.input_text(
                intro=intro,
                variable="author",
                default="Jane Developer",
                override_value=override_value,
            )
        elif len(options) > 1:
            # Add a line with all authors joined: E.g. 'Jane Developer & Joe Developer'
            # and a line with Other
            options.append(", ".join(options[:-1]) + f" & {options[-1]}")

        options.append("Other")

        # We want to use the input_text-function if override_value is provided or if the selected author is "Other"
        # However, since we don't want the select_option prompt if an override value is provided, we need to
        # initialise the author variable here.
        author = self.select_option(
            intro=(
                intro
                + "\n\n"
                + "We found these author names in the PEP621 formatted "
                + "'pyproject.toml'. Who do you want to be credited as the author of "
                + "this application?"
            ),
            variable="Author",
            options=options,
            default=None,
            override_value=None,
        )
        if author == "Other":
            author = self.input_text(
                intro="Who do you want to be credited as the author of this application?",
                variable="author",
                default="Jane Developer",
                override_value=None,
            )

        return author

    def input_email(self, author: str, bundle: str, override_value: str | None) -> str:
        """Ask about the author email, using hints from the pyproject.toml file if there
        are any.

        :returns: author email
        """
        default = self.make_author_email(author, bundle)
        default_source = "the author name and bundle"
        for author_info in self.pep621_data.get("authors", []):
            if author_info.get("name") == author and author_info.get("email"):
                default = author_info["email"]
                default_source = "the selected author name"

        intro = (
            "What email address should people use to contact the developers of this "
            "application? This might be your own email address, or a generic contact address "
            f"you set up specifically for this application. Based on {default_source}, "
            f"we believe it could be '{default}'."
        )

        author_email = self.input_text(
            intro=intro,
            variable="author's email",
            default=default,
            validator=self.validate_email,
            override_value=override_value,
        )

        return author_email

    def get_license_from_text(self, license_text: str) -> str:
        """Infer the license from the license file."""
        # The order here is quite important. If we have GPLvX+ after GPLvX, then it will
        # never be matched if the license text is GPLvX+, since it will already have
        # matched GPLvX. We search for MIT last, because words like PERMITTED and
        # LIMITED will generate a false match.

        hint_patterns = {
            "Apache Software License": ["Apache"],
            "BSD license": [
                "Redistribution and use in source and binary forms",
                "BSD",
            ],
            "GNU General Public License v2 or later (GPLv2+)": [
                "Free Software Foundation, either version 2 of the License",
                "GPLv2+",
            ],
            "GNU General Public License v2 (GPLv2)": [
                "version 2 of the GNU General Public License",
                "GPLv2",
            ],
            "GNU General Public License v3 or later (GPLv3+)": [
                "either version 3 of the License",
                "GPLv3+",
            ],
            "GNU General Public License v3 (GPLv3)": [
                "version 3 of the GNU General Public License",
                "GPLv3",
            ],
            "MIT license": [
                "Permission is hereby granted, free of charge",
                "MIT",
            ],
        }
        for hint, license_patterns in hint_patterns.items():
            for license_pattern in license_patterns:
                if license_pattern.lower() in license_text.lower():
                    return hint

        return "Other"

    def get_license_hint(self) -> tuple[str | None, str]:
        """Get hint for project license, either by reading pyproject.toml or the license
        file.

        :returns: The default value of the license - None if we cannot infer a license.
        :returns: The intro text to use when prompting the license
        """
        intro = "What license do you want to use for this project's code? "
        default = None

        # If there is license information in the pyproject.toml file, use that, otherwise check the license file
        if "text" in self.pep621_data.get("license", {}):
            default = self.get_license_from_text(self.pep621_data["license"]["text"])
            default_source = "the PEP621 formatted pyproject.toml"
        elif "file" in self.pep621_data.get("license", {}):
            license_text = (
                self.base_path / self.pep621_data["license"]["file"]
            ).read_text(encoding="utf-8")
            default = self.get_license_from_text(license_text)
            default_source = "the license file"
        elif (self.base_path / "LICENSE").exists():
            license_text = (self.base_path / "LICENSE").read_text(encoding="utf-8")
            default = self.get_license_from_text(license_text)
            default_source = "the license file"
        elif (self.base_path / "LICENCE").exists():
            license_text = (self.base_path / "LICENCE").read_text(encoding="utf-8")
            default = self.get_license_from_text(license_text)
            default_source = "the license file"
        else:
            return None, intro

        intro += f"\nBased on {default_source} we believe it is {default!r}."
        return default, intro

    def input_license(self, override_value: str | None) -> str:
        """Ask about the license, using hints from the pyproject.toml or license file if
        there are any.

        :returns: The project
        """
        default, intro = self.get_license_hint()
        options = [
            "BSD license",
            "MIT license",
            "Apache Software License",
            "GNU General Public License v2 (GPLv2)",
            "GNU General Public License v2 or later (GPLv2+)",
            "GNU General Public License v3 (GPLv3)",
            "GNU General Public License v3 or later (GPLv3+)",
            "Proprietary",
            "Other",
        ]

        return self.select_option(
            intro=intro,
            variable="Project License",
            options=options,
            default=default,
            override_value=override_value,
        )

    def build_app_context(self, project_overrides):
        """Ask the user for details about the app to be created.

        :returns: A context dictionary to be used in the cookiecutter project template.
        """
        app_name = self.input_app_name(
            override_value=project_overrides.pop("app_name", None)
        )
        formal_name = self.input_formal_name(
            app_name, override_value=project_overrides.pop("formal_name", None)
        )
        # The class name can be completely derived from the formal name.
        class_name = make_class_name(formal_name)
        # The module name can be completely derived from the app name.
        module_name = self.make_module_name(app_name)
        source_dir = self.input_source_dir(
            app_name,
            module_name,
            override_value=project_overrides.pop("source_dir", None),
        )
        test_source_dir = self.input_test_source_dir(
            module_name, override_value=project_overrides.pop("test_source_dir", None)
        )
        project_name = self.input_project_name(
            formal_name, override_value=project_overrides.pop("project_name", None)
        )
        description = self.input_description(
            override_value=project_overrides.pop("description", None)
        )
        url = self.input_url(app_name, project_overrides.pop("url", None))
        bundle = self.input_bundle(url, app_name, project_overrides.pop("bundle", None))
        author = self.input_author(override_value=project_overrides.pop("author", None))
        author_email = self.input_email(
            author, bundle, override_value=project_overrides.pop("author_email", None)
        )
        project_license = self.input_license(
            override_value=project_overrides.pop("license", None)
        )

        return {
            "formal_name": formal_name,
            "app_name": app_name,
            "class_name": class_name,
            "module_name": module_name,
            "project_name": project_name,
            "source_dir": source_dir,
            "test_source_dir": test_source_dir,
            "description": description,
            "author": author,
            "author_email": author_email,
            "bundle": bundle,
            "url": url,
            "license": project_license,
        }

    def build_gui_context(
        self,
        context: dict[str, str],
        project_overrides: dict[str, str],
    ) -> dict[str, str]:
        # We must set the GUI-framework to None here since the convert-command uses the new-command
        # template. This template includes dependencies for the GUI-frameworks. However, if a project
        # already is set up for a GUI-framework, then those dependencies should already be listed.
        # To prevent the same dependency being listed twice (once in the PEP621-section and once in the
        # briefcase-section), possibly with different versions, we set the GUI-framework to None here.
        return {"gui_framework": "None"}

    def merge_or_copy_pyproject(self, briefcase_config_file: Path) -> None:
        """Merge pyproject.toml file made by the cookiecutter with the one in the
        existing project.

        If the target directory doesn't have a pyproject.toml file, then the newly
        created will be copied

        :param briefcase_config_file: The path to the project created by the
            cookiecutter.
        """
        pyproject_file = self.base_path / "pyproject.toml"

        briefcase_pyproject = briefcase_config_file.read_text(encoding="utf-8")
        if pyproject_file.exists():
            pep621_pyproject = pyproject_file.read_text(encoding="utf-8")

            # The pyproject.toml file in the target directory has no briefcase keys, so it's
            # safe to copy-paste the text, and that way also keep formatting and comments.
            # We merge it this way to preserve comments in the original pyproject.toml file
            briefcase_comment = "# content below this line added by briefcase convert"
            merged_pyproject = (
                f"{pep621_pyproject}\n\n\n{briefcase_comment}\n{briefcase_pyproject}"
            )
        else:
            merged_pyproject = briefcase_pyproject

        with open(pyproject_file, "w", encoding="utf-8") as file:
            file.write(merged_pyproject)

    def migrate_necessary_files(self, project_dir, test_source_dir, module_name):
        """Copy and merge the necessary files from project_dir to the current base path.

        Will warn a LICENSE or CHANGELOG file is missing.

        :param project_dir: The path to the project created by the cookiecutter.
        :param test_source_dir: The path to the directory that should contain the test
            entry script
        """
        self.merge_or_copy_pyproject(project_dir / "pyproject.toml")

        # Copy license file if not already there
        license_file = self.pep621_data.get("license", {}).get("file")

        if license_file is None and not (self.base_path / "LICENSE").exists():
            self.logger.warning(
                f"\nLicense file not found in '{self.base_path}'. "
                "Briefcase will create a template 'LICENSE' file."
            )
            copy2(project_dir / "LICENSE", self.base_path / "LICENSE")

        # Copy changelog file
        changelog_file = self.base_path / "CHANGELOG"
        if not changelog_file.is_file():
            self.logger.warning(
                f"\nChangelog file not found in '{self.base_path}'. You should either "
                f"create a new '{self.base_path / 'CHANGELOG'}' file, or rename an "
                "already existing changelog file to 'CHANGELOG'."
            )

        # Copy tests or test entry script
        test_path = self.base_path / test_source_dir
        if test_path.exists():
            test_entry_script = project_dir / test_source_dir / f"{module_name}.py"
            copy2(
                test_entry_script,
                test_path / f"{module_name}.py",
            )
        else:
            copytree(project_dir / test_source_dir, test_path)

    def convert_app(
        self,
        tmp_path: Path,
        template: str | None = None,
        template_branch: str | None = None,
        project_overrides: dict[str, str] | None = None,
        **options,
    ) -> None:
        """Run the wizard in a temporary directory and copy the necessary files into the
        project.

        :param tmp_path: Temporary path that should contain the files generated by
            cookiecutter.
        :param template: The cookiecutter template to use.
        :param template_branch: The git branch that the template should use.
        """
        self.input.prompt()
        self.input.prompt("Let's setup an existing project as a Briefcase app!")
        context = self.build_context(project_overrides)
        self.prompt_divider()  # close the prompting section of output

        self.warn_unused_overrides(project_overrides)

        self.logger.info()
        self.logger.info(
            f"Generating required files to set up '{context['formal_name']}' with Briefcase",
            prefix=context["app_name"],
        )

        # Create the project files
        self.generate_template(
            template=(
                template
                if template
                else "https://github.com/beeware/briefcase-template"
            ),
            branch=template_branch,
            output_path=tmp_path,
            extra_context=context,
        )

        project_dir = tmp_path / context["app_name"]
        self.migrate_necessary_files(
            project_dir, context["test_source_dir"], context["module_name"]
        )

        self.logger.info(
            f"Converted existing application '{context['formal_name']}'",
            prefix=context["app_name"],
        )
        self.logger.info(
            """
To run your application, type:

    $ briefcase dev

"""
        )

    def validate_pyproject_file(self) -> None:
        """Cannot setup new app if it already has briefcase settings in pyproject."""
        if not (self.base_path / "pyproject.toml").exists():
            raise BriefcaseCommandError(
                "Cannot automatically set up Briefcase for a project without a "
                "'pyproject.toml' file."
            )

        with open(self.base_path / "pyproject.toml", "rb") as file:
            pyproject = tomllib.load(file)

        if "tool" in pyproject and "briefcase" in pyproject["tool"]:
            raise BriefcaseCommandError(
                "Your 'pyproject.toml' file already contains a Briefcase configuration."
            )

    def __call__(
        self,
        template: str | None = None,
        template_branch: str | None = None,
        project_overrides: list[str] = None,
        **options,
    ):
        # Confirm host compatibility, and that all required tools are available.
        # There are no apps, so finalize() will be a no op on app configurations.
        self.finalize()

        self.validate_pyproject_file()

        # Setup the app for briefcase
        with TemporaryDirectory() as tmp_path:
            tmp_path = Path(tmp_path)
            self.convert_app(
                tmp_path=tmp_path,
                template=template,
                template_branch=template_branch,
                project_overrides=parse_project_overrides(project_overrides),
                **options,
            )



from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.platforms.macOS import SigningIdentity
from briefcase.platforms.macOS.app import macOSAppBuildCommand


@pytest.fixture
def build_command(tmp_path):
    command = macOSAppBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )

    command.select_identity = mock.MagicMock()
    command.sign_app = mock.MagicMock()
    command.sign_file = mock.MagicMock()

    return command


@pytest.mark.parametrize("universal_build", [True, False])
@pytest.mark.parametrize("pre_existing", [True, False])
@pytest.mark.parametrize("console_app", [True, False])
def test_build_app(
    build_command,
    first_app_with_binaries,
    console_app,
    universal_build,
    pre_existing,
    tmp_path,
):
    """A macOS App is ad-hoc signed as part of the build process."""
    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    first_app_with_binaries.universal_build = universal_build
    first_app_with_binaries.console_app = console_app
    build_command.tools.host_arch = "gothic"

    exec_path = bundle_path / "First App.app/Contents/MacOS"
    if not pre_existing:
        # If this not a pre-existing app, the stub has the original name
        (exec_path / "First App").rename(exec_path / "Stub")

    # Mock the thin command so we can confirm if it was invoked.
    build_command.ensure_thin_binary = mock.Mock()

    # Build the app
    build_command.build_app(first_app_with_binaries, test_mode=False)

    # The stub binary has been renamed
    assert not (exec_path / "Stub").is_file()
    assert (exec_path / "First App").is_file()

    # Only thin if this is a non-universal app
    if universal_build:
        build_command.ensure_thin_binary.assert_not_called()
    else:
        build_command.ensure_thin_binary.assert_called_once_with(
            exec_path / "First App",
            arch="gothic",
        )

    # A request has been made to sign the app
    build_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=SigningIdentity(),
    )

    # No request to select a signing identity was made
    build_command.select_identity.assert_not_called()

    # No attempt was made to sign a specific file;
    # This ignores the calls that would have been made transitively
    # by calling sign_app()
    build_command.sign_file.assert_not_called()



import pytest


@pytest.mark.parametrize(
    "formal_name, candidate",
    [
        ("Hello World", "helloworld"),
        ("Hello World!", "helloworld"),
        ("Hello! World", "helloworld"),
        ("Hello-World", "helloworld"),
        # Internationalized names that can be unicode-simplified
        ("Hallo Vögel", "hallovogel"),
        ("Bonjour Garçon", "bonjourgarcon"),
        # Internationalized names that cannot be unicode-simplified
        ("你好 世界", "myapp"),
    ],
)
def test_make_app_name(new_command, formal_name, candidate):
    """A formal name can be converted into a valid app name."""
    app_name = new_command.make_app_name(formal_name)
    assert app_name == candidate
    # Double check - the app name passes the validity check.
    assert new_command.validate_app_name(app_name)



import pytest

from briefcase.platforms.android.gradle import android_log_clean_filter


@pytest.mark.parametrize(
    "original, filtered",
    [
        # Start of app log
        (
            "--------- beginning of main",
            ("--------- beginning of main", False),
        ),
        # System messages log
        (
            "D/libEGL  : loaded /vendor/lib64/egl/libEGL_emulation.so",
            (
                "loaded /vendor/lib64/egl/libEGL_emulation.so",
                False,
            ),
        ),
        (
            "\x1b[32mD/libEGL  : loaded /vendor/lib64/egl/libEGL_emulation.so\x1b[0m",
            (
                "loaded /vendor/lib64/egl/libEGL_emulation.so",
                False,
            ),
        ),
        (
            "D/stdio   : Could not find platform independent libraries <prefix>",
            ("Could not find platform independent libraries <prefix>", False),
        ),
        (
            "D/MainActivity: onStart() start",
            ("onStart() start", False),
        ),
        (
            "\x1b[32mD/MainActivity: onStart() start\x1b[0m",
            ("onStart() start", False),
        ),
        # Python App messages
        (
            "I/python.stdout: Python app launched & stored in Android Activity class",
            ("Python app launched & stored in Android Activity class", True),
        ),
        (
            "\x1b[32mI/python.stdout: Python app launched & stored in Android Activity class\x1b[0m",
            ("Python app launched & stored in Android Activity class", True),
        ),
        (
            "I/python.stdout: ",
            ("", True),
        ),
        (
            "\x1b[32mI/python.stdout: \x1b[0m",
            ("", True),
        ),
        (
            "\x1b[32m\x1b[98mI/python.stdout: \x1b[32m\x1b[0m",
            ("", True),
        ),
        (
            "\x1b[32m\x1b[98mI/python.stdout: this is colored output\x1b[32m\x1b[0m",
            ("this is colored output", True),
        ),
        (
            "I/python.stderr: test_case (tests.foobar.test_other.TestOtherMethods)",
            ("test_case (tests.foobar.test_other.TestOtherMethods)", True),
        ),
        (
            "\x1b[32mI/python.stderr: test_case (tests.foobar.test_other.TestOtherMethods)\x1b[0m",
            ("test_case (tests.foobar.test_other.TestOtherMethods)", True),
        ),
        (
            "I/python.stderr: ",
            ("", True),
        ),
        (
            "\x1b[32mI/python.stderr: \x1b[0m",
            ("", True),
        ),
        # Unknown content
        (
            "This doesn't match the regex",
            ("This doesn't match the regex", False),
        ),
        (
            "\x1b[33mThis doesn't match the regex\x1b[33m",
            ("\x1b[33mThis doesn't match the regex\x1b[33m", False),
        ),
    ],
)
def test_filter(original, filtered):
    assert android_log_clean_filter(original) == filtered



from briefcase.integrations.file import File
from briefcase.integrations.subprocess import Subprocess


def test_base_tools_exist(base_command):
    """Ensure default tools are always available."""
    assert isinstance(base_command.tools.subprocess, Subprocess)
    assert isinstance(base_command.tools.file, File)

    base_command.verify_tools()



import pytest

from briefcase.exceptions import UnsupportedHostError
from briefcase.integrations.xcode import Xcode


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.xcode = "tool"

    tool = Xcode.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.xcode


@pytest.mark.parametrize("host_os", ["Linux", "Windows", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{Xcode.name} is not supported on {host_os}",
    ):
        Xcode.verify(mock_tools)



import shutil
import subprocess
import sys
from unittest import mock

if sys.version_info >= (3, 11):  # pragma: no-cover-if-lt-py311
    import tomllib
else:  # pragma: no-cover-if-gte-py311
    import tomli as tomllib

import pytest

from briefcase.console import Console, Log, LogLevel
from briefcase.exceptions import BriefcaseCommandError, BriefcaseConfigError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.web.static import StaticWebBuildCommand

from ....utils import create_wheel


@pytest.fixture
def build_command(tmp_path):
    command = StaticWebBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command.tools.shutil = mock.MagicMock(spec_set=shutil)

    return command


@pytest.mark.parametrize("logging_level", [LogLevel.INFO, LogLevel.DEEP_DEBUG])
def test_build_app(build_command, first_app_generated, logging_level, tmp_path):
    """An app can be built."""
    # Configure logging level
    build_command.logger.verbosity = logging_level

    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    # Invoking build will create wheels as a side effect.
    def mock_run(*args, **kwargs):
        if args[0][5] == "wheel":
            create_wheel(
                bundle_path / "www/static/wheels",
                "first_app",
                extra_content=[
                    ("dependency/static/style.css", "span { margin: 10px; }\n"),
                ],
            ),
        elif args[0][5] == "pip":
            create_wheel(
                bundle_path / "www/static/wheels",
                "dependency",
                extra_content=[
                    ("dependency/static/style.css", "div { margin: 10px; }\n"),
                ],
            ),
            create_wheel(
                bundle_path / "www/static/wheels",
                "other",
                extra_content=[
                    ("other/static/style.css", "div { padding: 10px; }\n"),
                ],
            ),
        else:
            raise ValueError("Unknown command")

    build_command.tools.subprocess.run.side_effect = mock_run

    # Mock the side effect of invoking shutil
    build_command.tools.shutil.rmtree.side_effect = lambda *args: shutil.rmtree(
        bundle_path / "www/static/wheels"
    )

    # Build the web app.
    build_command.build_app(first_app_generated)

    # The old wheel folder was removed
    build_command.tools.shutil.rmtree.assert_called_once_with(
        bundle_path / "www/static/wheels"
    )

    # `wheel pack` and `pip wheel` was invoked
    assert build_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "wheel",
                "pack",
                bundle_path / "app",
                "--dest-dir",
                bundle_path / "www/static/wheels",
            ],
            check=True,
            encoding="UTF-8",
        ),
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "wheel",
                "--wheel-dir",
                bundle_path / "www/static/wheels",
                "-r",
                bundle_path / "requirements.txt",
            ]
            + (["-vv"] if logging_level == LogLevel.DEEP_DEBUG else []),
            check=True,
            encoding="UTF-8",
        ),
    ]

    # Pyscript.toml has been written
    with (bundle_path / "www/pyscript.toml").open("rb") as f:
        assert tomllib.load(f) == {
            "name": "First App",
            "description": "The first simple app \\ demonstration",
            "version": "0.0.1",
            "splashscreen": {"autoclose": True},
            "terminal": False,
            "packages": [
                "/static/wheels/dependency-1.2.3-py3-none-any.whl",
                "/static/wheels/first_app-1.2.3-py3-none-any.whl",
                "/static/wheels/other-1.2.3-py3-none-any.whl",
            ],
        }

    # briefcase.css has been appended
    with (bundle_path / "www/static/css/briefcase.css").open(encoding="utf-8") as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "",
                    "#pyconsole {",
                    "  display: None;",
                    "}",
                    "/*******************************************************************",
                    " ******************** Wheel contributed styles ********************/",
                    "",
                    "/*******************************************************",
                    " * dependency 1.2.3::style.css",
                    " *******************************************************/",
                    "",
                    "div { margin: 10px; }",
                    "",
                    "/*******************************************************",
                    " * first_app 1.2.3::style.css",
                    " *******************************************************/",
                    "",
                    "span { margin: 10px; }",
                    "",
                    "/*******************************************************",
                    " * other 1.2.3::style.css",
                    " *******************************************************/",
                    "",
                    "div { padding: 10px; }",
                ]
            )
            + "\n"
        )


def test_build_app_custom_pyscript_toml(build_command, first_app_generated, tmp_path):
    """An app with extra pyscript.toml content can be written."""
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    first_app_generated.extra_pyscript_toml_content = (
        "\n".join(
            [
                # A custom top level key
                'something = "custom"',
                # A key overwriting something in the main namespace
                'packages = ["something-custom"]',
                # A table overwriting an existing value
                "[splashscreen]",
                "wiggle = false",
                # A custom array of tables
                "[[runtimes]]",
                'src = "https://example.com/pyodide.js"',
            ]
        )
        + "\n"
    )

    # Mock the side effect of invoking shutil
    build_command.tools.shutil.rmtree.side_effect = lambda *args: shutil.rmtree(
        bundle_path / "www/static/wheels"
    )

    # Build the web app.
    build_command.build_app(first_app_generated)

    # Pyscript.toml has been written
    with (bundle_path / "www/pyscript.toml").open("rb") as f:
        assert tomllib.load(f) == {
            "name": "First App",
            "description": "The first simple app \\ demonstration",
            "version": "0.0.1",
            "something": "custom",
            "splashscreen": {"wiggle": False},
            "terminal": False,
            "packages": ["something-custom"],
            "runtimes": [
                {"src": "https://example.com/pyodide.js"},
            ],
        }


def test_build_app_invalid_custom_pyscript_toml(
    build_command, first_app_generated, tmp_path
):
    """An app with invalid extra pyscript.toml content raises an error."""
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    first_app_generated.extra_pyscript_toml_content = "This isn't toml"

    # Mock the side effect of invoking shutil
    build_command.tools.shutil.rmtree.side_effect = lambda *args: shutil.rmtree(
        bundle_path / "www/static/wheels"
    )

    # Building the web app raises an error
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Briefcase configuration error: Extra pyscript.toml content isn't valid TOML: Expected",
    ):
        build_command.build_app(first_app_generated)


def test_build_app_missing_wheel_dir(build_command, first_app_generated, tmp_path):
    """If the template is corrupted, the wheels folder is still created."""
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    # Remove the wheels folder
    shutil.rmtree(bundle_path / "www/static/wheels")

    # Build the web app.
    build_command.build_app(first_app_generated)

    # No attempt was made to delete the folder
    build_command.tools.shutil.rmtree.assert_not_called()

    # `wheel pack` and `pip wheel` was invoked
    assert build_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "wheel",
                "pack",
                bundle_path / "app",
                "--dest-dir",
                bundle_path / "www/static/wheels",
            ],
            check=True,
            encoding="UTF-8",
        ),
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "wheel",
                "--wheel-dir",
                bundle_path / "www/static/wheels",
                "-r",
                bundle_path / "requirements.txt",
            ],
            check=True,
            encoding="UTF-8",
        ),
    ]

    # Wheels folder exists
    assert (bundle_path / "www/static/wheels").is_dir()

    # Pyscript.toml has been written
    assert (bundle_path / "www/pyscript.toml").exists()

    # briefcase.css has been appended
    # (just check for existence; a full check is done in other tests)
    assert (bundle_path / "www/static/css/briefcase.css").exists()


def test_build_app_no_requirements(build_command, first_app_generated, tmp_path):
    """An app with no requirements can be built."""
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    # Invoking build will create wheels as a side effect.
    def mock_run(*args, **kwargs):
        if args[0][5] == "wheel":
            create_wheel(
                bundle_path / "www/static/wheels",
                "first_app",
                extra_content=[
                    ("dependency/static/style.css", "span { margin: 10px; }\n"),
                ],
            ),
        elif args[0][5] == "pip":
            pass
        else:
            raise ValueError("Unknown command")

    build_command.tools.subprocess.run.side_effect = mock_run

    # Mock the side effect of invoking shutil
    build_command.tools.shutil.rmtree.side_effect = lambda *args: shutil.rmtree(
        bundle_path / "www/static/wheels"
    )

    # Build the web app.
    build_command.build_app(first_app_generated)

    # The old wheel folder was removed
    build_command.tools.shutil.rmtree.assert_called_once_with(
        bundle_path / "www/static/wheels"
    )

    # `wheel pack` and `pip wheel` was invoked
    assert build_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "wheel",
                "pack",
                bundle_path / "app",
                "--dest-dir",
                bundle_path / "www/static/wheels",
            ],
            check=True,
            encoding="UTF-8",
        ),
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "wheel",
                "--wheel-dir",
                bundle_path / "www/static/wheels",
                "-r",
                bundle_path / "requirements.txt",
            ],
            check=True,
            encoding="UTF-8",
        ),
    ]

    # Pyscript.toml has been written
    with (bundle_path / "www/pyscript.toml").open("rb") as f:
        assert tomllib.load(f) == {
            "name": "First App",
            "description": "The first simple app \\ demonstration",
            "version": "0.0.1",
            "splashscreen": {"autoclose": True},
            "terminal": False,
            "packages": [
                "/static/wheels/first_app-1.2.3-py3-none-any.whl",
            ],
        }

    # briefcase.css has been appended
    with (bundle_path / "www/static/css/briefcase.css").open(encoding="utf-8") as f:
        assert (
            f.read()
            == "\n".join(
                [
                    "",
                    "#pyconsole {",
                    "  display: None;",
                    "}",
                    "/*******************************************************************",
                    " ******************** Wheel contributed styles ********************/",
                    "",
                    "/*******************************************************",
                    " * first_app 1.2.3::style.css",
                    " *******************************************************/",
                    "",
                    "span { margin: 10px; }",
                ]
            )
            + "\n"
        )


def test_app_package_fail(build_command, first_app_generated, tmp_path):
    """If the app can't be packaged as a wheel,an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    # Mock the side effect of invoking shutil
    build_command.tools.shutil.rmtree.side_effect = lambda *args: shutil.rmtree(
        bundle_path / "www/static/wheels"
    )

    # Mock the side effect of a successful package build, but failure
    # downloading requirements
    build_command.tools.subprocess.run.side_effect = [
        subprocess.CalledProcessError(cmd=["wheel"], returncode=1),
    ]

    # Build the web app.
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to build wheel for app 'first-app'",
    ):
        build_command.build_app(first_app_generated)

    # The old wheel folder was removed
    build_command.tools.shutil.rmtree.assert_called_once_with(
        bundle_path / "www/static/wheels"
    )

    # `wheel pack` was invoked
    assert build_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "wheel",
                "pack",
                bundle_path / "app",
                "--dest-dir",
                bundle_path / "www/static/wheels",
            ],
            check=True,
            encoding="UTF-8",
        ),
    ]

    # Wheels folder still exists
    assert (bundle_path / "www/static/wheels").is_dir()

    # Pyscript.toml was not written
    assert not (bundle_path / "www/pyscript.toml").exists()


def test_dependency_fail(build_command, first_app_generated, tmp_path):
    """If dependencies can't be downloaded, an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/web/static"

    # Mock the side effect of invoking shutil
    build_command.tools.shutil.rmtree.side_effect = lambda *args: shutil.rmtree(
        bundle_path / "www/static/wheels"
    )

    # Mock the side effect of a successful package build, but failure
    # downloading requirements
    build_command.tools.subprocess.run.side_effect = [
        None,
        subprocess.CalledProcessError(cmd=["pip", "wheel"], returncode=1),
    ]

    # Build the web app.
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to install requirements for app 'first-app'",
    ):
        build_command.build_app(first_app_generated)

    # The old wheel folder was removed
    build_command.tools.shutil.rmtree.assert_called_once_with(
        bundle_path / "www/static/wheels"
    )

    # `wheel pack` and `pip wheel` was invoked
    assert build_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "wheel",
                "pack",
                bundle_path / "app",
                "--dest-dir",
                bundle_path / "www/static/wheels",
            ],
            check=True,
            encoding="UTF-8",
        ),
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "wheel",
                "--wheel-dir",
                bundle_path / "www/static/wheels",
                "-r",
                bundle_path / "requirements.txt",
            ],
            check=True,
            encoding="UTF-8",
        ),
    ]

    # Wheels folder still exists
    assert (bundle_path / "www/static/wheels").is_dir()

    # Pyscript.toml was not written
    assert not (bundle_path / "www/pyscript.toml").exists()



import os
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.docker import Docker, DockerAppContext
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.appimage import LinuxAppImageOpenCommand

from ...utils import create_file


@pytest.fixture
def open_command(tmp_path):
    command = LinuxAppImageOpenCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.os = MagicMock(spec_set=os)

    # Mock x86_64 for linuxdeploy verification
    command.tools.host_arch = "x86_64"

    # Store the underlying subprocess instance
    command._subprocess = MagicMock(spec_set=Subprocess)
    command.tools.subprocess = command._subprocess

    # Mock existence of linuxdeploy in tools
    create_file(
        command.tools.base_path / "linuxdeploy-x86_64.AppImage",
        content="",
    )

    return command


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows paths aren't converted in Docker context",
)
def test_open_docker(open_command, first_app_config, tmp_path, monkeypatch):
    """Open starts a docker session by default."""

    # Enable docker
    open_command.use_docker = True
    open_command.extra_docker_build_args = []

    # Provide Docker
    monkeypatch.setattr(
        Docker, "_is_user_mapping_enabled", MagicMock(return_value=True)
    )
    open_command.tools.docker = Docker(tools=open_command.tools)
    # Provide Docker app context
    open_command.tools[first_app_config].app_context = DockerAppContext(
        tools=open_command.tools,
        app=first_app_config,
    )
    open_command.tools[first_app_config].app_context.prepare(
        image_tag=f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/linux/appimage/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/linux/appimage",
        host_data_path=tmp_path / "briefcase",
        python_version=f"3.{sys.version_info.minor}",
    )
    # Clear out any calls recorded during preparation
    open_command._subprocess.run.reset_mock()

    # Create the desktop file that would be in the project folder.
    create_file(
        open_command.project_path(first_app_config)
        / "First App.AppDir/com.example.firstapp.desktop",
        "FreeDesktop file",
    )

    # Open the app
    open_command.open_app(first_app_config)

    # The docker session was started
    open_command._subprocess.run.assert_called_once_with(
        args=[
            "docker",
            "run",
            "--rm",
            "-it",
            "--volume",
            f"{open_command.base_path}/build/first-app/linux/appimage:/app:z",
            "--volume",
            f"{open_command.data_path}:/briefcase:z",
            f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
        ],
        stream_output=False,
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.skipif(sys.platform != "linux", reason="Linux specific test")
def test_open_no_docker_linux(open_command, first_app_config, tmp_path):
    """On Linux, Open runs `xdg-open` on the project folder if we specify --no-
    docker."""
    # Create the desktop file that would be in the project folder.
    create_file(
        open_command.project_path(first_app_config)
        / "First App.AppDir/com.example.firstapp.desktop",
        "FreeDesktop file",
    )

    # Disable docker
    open_command.use_docker = False

    open_command(first_app_config)

    open_command.tools.subprocess.Popen.assert_called_once_with(
        [
            "xdg-open",
            tmp_path / "base_path/build/first-app/linux/appimage",
        ]
    )


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
def test_open_no_docker_macOS(open_command, first_app_config, tmp_path):
    """On macOS, Open runs `open` on the project folder if we specify --no-docker."""
    # Create the desktop file that would be in the project folder.
    create_file(
        open_command.project_path(first_app_config)
        / "First App.AppDir/com.example.firstapp.desktop",
        "FreeDesktop file",
    )

    # Disable docker
    open_command.use_docker = False

    open_command(first_app_config)

    open_command.tools.subprocess.Popen.assert_called_once_with(
        [
            "open",
            tmp_path / "base_path/build/first-app/linux/appimage",
        ]
    )



import os
import platform
import shutil
import stat
from pathlib import Path
from unittest import mock

import pytest
import requests
import requests.exceptions
from urllib3._collections import HTTPHeaderDict

from briefcase.exceptions import (
    BadNetworkResourceError,
    MissingNetworkResourceError,
    NetworkFailure,
)
from briefcase.integrations.base import ToolCache

TEMPORARY_DOWNLOAD_FILE_SUFFIX = ".download"


@pytest.fixture
def mock_tools(mock_tools) -> ToolCache:
    mock_tools.requests = mock.MagicMock(spec_set=requests)
    # Restore move so the temporary file can be moved after downloaded
    mock_tools.shutil.move = mock.MagicMock(wraps=shutil.move)
    return mock_tools


@pytest.fixture
def file_perms() -> int:
    """The expected permissions for the downloaded file.

    Since umask can vary on different systems, it is updated to a known value and reset
    after the test finishes.

    On Windows, the umask seems to always be zero and chmod doesn't really do anything
    anyway.
    """
    if platform.system() == "Windows":
        yield 0o666
    else:
        orig_umask = os.umask(0o077)
        yield 0o600
        os.umask(orig_umask)


@pytest.mark.parametrize(
    "url,content_disposition",
    [
        # A `None` value for `content_disposition` means we skip the header.
        # Other values are passed through as HTTP header values.
        ("https://example.com/path/to/something.zip", None),
        # Ensure empty header is ignored.
        ("https://example.com/path/to/something.zip", ""),
        # Paradigmatic case for Content-Disposition: attachment.
        (
            "https://example.com/path/to/irrelevant.zip",
            "attachment; filename=something.zip",
        ),
        # Ensure extra parameters are ignored.
        (
            "https://example.com/path/to/irrelevant.zip",
            "attachment; filename=something.zip; ignored=okay",
        ),
        # Ensure garbage headers are ignored.
        ("https://example.com/path/to/something.zip", "garbage"),
        # Ensure we respect unusual quoting & case permitted by RFC 6266.
        (
            "https://example.com/path/to/irrelevant.zip",
            'ATTACHment; filename=    "something.zip"',
        ),
        # Ensure we use filename=, even if filename*= is provided. This makes us a
        # "legacy user agent" in the terms of RFC 6266, for our own simplicity.
        (
            "https://example.com/path/to/irrelevant.zip",
            'attachment; filename="something.zip"; filename*=utf-8'
            "''%e2%82%ac%20rates",
        ),
    ],
)
def test_new_download_oneshot(mock_tools, file_perms, url, content_disposition):
    response = mock.MagicMock(spec=requests.Response)
    response.url = url
    response.status_code = 200
    response.headers = mock.Mock(
        wraps=HTTPHeaderDict(
            {
                "content-disposition": content_disposition,
            }
            if content_disposition is not None
            else {}
        )
    )
    response.content = b"all content"
    mock_tools.requests.get.return_value = response

    # Download the file
    filename = mock_tools.file.download(
        url="https://example.com/support?useful=Yes",
        download_path=mock_tools.base_path / "downloads",
    )

    # requests.get has been invoked, but content isn't iterated
    mock_tools.requests.get.assert_called_with(
        "https://example.com/support?useful=Yes",
        stream=True,
    )
    response.headers.get.assert_called_with("content-length")
    response.iter_content.assert_not_called()

    # The filename is derived from the URL or header
    assert filename == mock_tools.base_path / "downloads/something.zip"

    # Temporary file was created in download dir and was renamed
    temp_filename = Path(mock_tools.shutil.move.call_args_list[0].args[0])
    assert temp_filename.parent == mock_tools.base_path / "downloads"
    assert temp_filename.name.startswith("something.zip.")
    assert temp_filename.name.endswith(TEMPORARY_DOWNLOAD_FILE_SUFFIX)
    mock_tools.shutil.move.assert_called_with(str(temp_filename), filename)

    # File permissions were set
    mock_tools.os.chmod.assert_called_with(filename, file_perms)
    assert stat.S_IMODE(os.stat(filename).st_mode) == file_perms

    # Attempt to delete Temporary file was made
    mock_tools.os.remove.assert_called_with(str(temp_filename))

    # File content is as expected
    with (mock_tools.base_path / "downloads/something.zip").open(encoding="utf-8") as f:
        assert f.read() == "all content"


def test_new_download_chunked(mock_tools, file_perms):
    response = mock.MagicMock()
    response.url = "https://example.com/path/to/something.zip"
    response.status_code = 200
    response.headers.get.return_value = "24"
    response.iter_content.return_value = iter(
        [
            b"chunk-1;",
            b"chunk-2;",
            b"chunk-3;",
        ]
    )
    mock_tools.requests.get.return_value = response

    # Download the file
    filename = mock_tools.file.download(
        url="https://example.com/support?useful=Yes",
        download_path=mock_tools.base_path,
    )

    # requests.get has been invoked, and content is chunked.
    mock_tools.requests.get.assert_called_with(
        "https://example.com/support?useful=Yes",
        stream=True,
    )
    response.headers.get.assert_called_with("content-length")
    response.iter_content.assert_called_once_with(chunk_size=1048576)

    # The filename is derived from the URL
    assert filename == mock_tools.base_path / "something.zip"

    # Temporary file was created in download dir and was renamed
    temp_filename = Path(mock_tools.shutil.move.call_args_list[0].args[0])
    assert temp_filename.parent == mock_tools.base_path
    assert temp_filename.name.startswith("something.zip.")
    assert temp_filename.name.endswith(TEMPORARY_DOWNLOAD_FILE_SUFFIX)
    mock_tools.shutil.move.assert_called_with(str(temp_filename), filename)

    # File permissions were set
    mock_tools.os.chmod.assert_called_with(filename, file_perms)
    assert stat.S_IMODE(os.stat(filename).st_mode) == file_perms

    # Attempt to delete Temporary file was made
    mock_tools.os.remove.assert_called_with(str(temp_filename))

    # The downloaded file exists, and content is as expected
    assert filename.exists()
    with (mock_tools.base_path / "something.zip").open(encoding="utf-8") as f:
        assert f.read() == "chunk-1;chunk-2;chunk-3;"


def test_already_downloaded(mock_tools):
    # Create an existing file
    existing_file = mock_tools.base_path / "something.zip"
    with existing_file.open("w", encoding="utf-8") as f:
        f.write("existing content")

    response = mock.MagicMock()
    response.headers.get.return_value = ""
    response.url = "https://example.com/path/to/something.zip"
    response.status_code = 200
    mock_tools.requests.get.return_value = response

    # Download the file
    filename = mock_tools.file.download(
        url="https://example.com/support?useful=Yes",
        download_path=mock_tools.base_path,
    )

    # The GET request will have been made
    mock_tools.requests.get.assert_called_with(
        "https://example.com/support?useful=Yes",
        stream=True,
    )

    # The request's Content-Disposition header is consumed to
    # examine the filename; the request is abandoned before
    # any other headers are read.
    response.headers.get.assert_called_once_with("Content-Disposition")

    # but the file existed, so the method returns
    assert filename == existing_file
    assert filename.exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_missing_resource(mock_tools):
    response = mock.MagicMock()
    response.status_code = 404

    mock_tools.requests.get.return_value = response

    # Download the file
    with pytest.raises(MissingNetworkResourceError):
        mock_tools.file.download(
            url="https://example.com/something.zip?useful=Yes",
            download_path=mock_tools.base_path,
        )

    # requests.get has been invoked, but nothing else.
    mock_tools.requests.get.assert_called_with(
        "https://example.com/something.zip?useful=Yes",
        stream=True,
    )
    response.headers.get.assert_not_called()

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_bad_resource(mock_tools):
    response = mock.MagicMock()
    response.status_code = 500

    mock_tools.requests.get.return_value = response

    # Download the file
    with pytest.raises(BadNetworkResourceError):
        mock_tools.file.download(
            url="https://example.com/something.zip?useful=Yes",
            download_path=mock_tools.base_path,
        )

    # requests.get has been invoked, but nothing else.
    mock_tools.requests.get.assert_called_with(
        "https://example.com/something.zip?useful=Yes",
        stream=True,
    )
    response.headers.get.assert_not_called()

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_get_connection_error(mock_tools):
    """NetworkFailure raises if requests.get() errors."""
    mock_tools.requests.get.side_effect = requests.exceptions.ConnectionError

    # Download the file
    with pytest.raises(
        NetworkFailure,
        match=r"Unable to download https\:\/\/example.com\/something\.zip\?useful=Yes",
    ):
        mock_tools.file.download(
            url="https://example.com/something.zip?useful=Yes",
            download_path=mock_tools.base_path,
        )

    # requests.get has been invoked, but nothing else.
    mock_tools.requests.get.assert_called_with(
        "https://example.com/something.zip?useful=Yes",
        stream=True,
    )

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_iter_content_connection_error(mock_tools):
    """NetworkFailure raised if response.iter_content() errors."""
    response = mock.MagicMock(spec=requests.Response)
    response.url = "https://example.com/something.zip?useful=Yes"
    response.headers = mock.Mock(wraps=HTTPHeaderDict({"content-length": "100"}))
    response.status_code = 200
    response.iter_content.side_effect = requests.exceptions.ConnectionError
    mock_tools.requests.get.return_value = response

    # Download the file
    with pytest.raises(NetworkFailure, match="Unable to download something.zip"):
        mock_tools.file.download(
            url="https://example.com/something.zip?useful=Yes",
            download_path=mock_tools.base_path,
        )

    # requests.get has been invoked, but nothing else.
    mock_tools.requests.get.assert_called_with(
        "https://example.com/something.zip?useful=Yes",
        stream=True,
    )
    response.headers.get.assert_called_with("content-length")

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not moved
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()

    # Temporary file was created and named appropriately and then deleted
    temp_filename = Path(mock_tools.os.remove.call_args_list[0].args[0])
    assert temp_filename.parent == mock_tools.base_path
    assert temp_filename.name.startswith("something.zip.")
    assert temp_filename.name.endswith(TEMPORARY_DOWNLOAD_FILE_SUFFIX)
    mock_tools.os.remove.assert_called_with(str(temp_filename))


def test_content_connection_error(mock_tools):
    """NetworkFailure raised if response.content errors."""
    response = mock.MagicMock(spec=requests.Response)
    response.url = "https://example.com/something.zip?useful=Yes"
    response.headers = mock.Mock(wraps=HTTPHeaderDict())
    response.status_code = 200
    type(response).content = mock.PropertyMock(
        side_effect=requests.exceptions.ConnectionError
    )
    mock_tools.requests.get.return_value = response

    # Download the file
    with pytest.raises(NetworkFailure, match="Unable to download something.zip"):
        mock_tools.file.download(
            url="https://example.com/something.zip?useful=Yes",
            download_path=mock_tools.base_path,
        )

    # requests.get has been invoked, but nothing else.
    mock_tools.requests.get.assert_called_with(
        "https://example.com/something.zip?useful=Yes",
        stream=True,
    )
    response.headers.get.assert_called_with("content-length")

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not moved but it was deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()

    # Temporary file was created and named appropriately and then deleted
    temp_filename = Path(mock_tools.os.remove.call_args_list[0].args[0])
    assert temp_filename.parent == mock_tools.base_path
    assert temp_filename.name.startswith("something.zip.")
    assert temp_filename.name.endswith(TEMPORARY_DOWNLOAD_FILE_SUFFIX)
    mock_tools.os.remove.assert_called_with(str(temp_filename))



from __future__ import annotations

import gzip
import re
import subprocess
import sys
from pathlib import Path

from briefcase.commands import (
    BuildCommand,
    CreateCommand,
    PackageCommand,
    PublishCommand,
    RunCommand,
    UpdateCommand,
)
from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.integrations.docker import Docker, DockerAppContext
from briefcase.integrations.subprocess import NativeAppContext
from briefcase.platforms.linux import (
    ARCH,
    DEBIAN,
    RHEL,
    SUSE,
    DockerOpenCommand,
    LinuxMixin,
    LocalRequirementsMixin,
    parse_freedesktop_os_release,
)


class LinuxSystemPassiveMixin(LinuxMixin):
    # The Passive mixin honors the Docker options, but doesn't try to verify
    # Docker exists. It is used by commands that are "passive" from the
    # perspective of the build system (e.g., Run).
    output_format = "system"
    supported_host_os = {"Darwin", "Linux"}
    supported_host_os_reason = (
        "Linux system projects can only be built on Linux, or on macOS using Docker."
    )

    @property
    def use_docker(self):
        # The system backend doesn't have a literal "--use-docker" option, but
        # `use_docker` is a useful flag for shared logic purposes, so evaluate
        # what "use docker" means in terms of target_image.
        return bool(self.target_image)

    def add_options(self, parser):
        super().add_options(parser)
        parser.add_argument(
            "--target",
            dest="target",
            help="Docker base image tag for the distribution to target for the build (e.g., `ubuntu:jammy`)",
            required=False,
        )
        parser.add_argument(
            "--Xdocker-build",
            action="append",
            dest="extra_docker_build_args",
            help="Additional arguments to use when building the Docker image",
            required=False,
        )

    def parse_options(self, extra):
        """Extract the target_image option."""
        options, overrides = super().parse_options(extra)
        self.target_image = options.pop("target")
        self.extra_docker_build_args = options.pop("extra_docker_build_args")

        return options, overrides

    def build_path(self, app):
        # Override the default build path to use the vendor name,
        # rather than "linux"
        return self.base_path / "build" / app.app_name / app.target_vendor

    def bundle_path(self, app):
        # Override the default bundle path to use the codename,
        # rather than "system"
        return self.build_path(app) / app.target_codename

    def project_path(self, app):
        return self.bundle_path(app) / f"{app.app_name}-{app.version}"

    def binary_path(self, app):
        return self.project_path(app) / "usr/bin" / app.app_name

    def rpm_tag(self, app):
        if app.target_vendor == "fedora":
            return f"fc{app.target_codename}"
        else:
            return f"el{app.target_codename}"

    def target_glibc_version(self, app):
        target_glibc = self.tools.os.confstr("CS_GNU_LIBC_VERSION").split()[1]
        return target_glibc

    def app_python_version_tag(self, app):
        # Use the version of Python that was used to run Briefcase.
        return self.python_version_tag

    def platform_freedesktop_info(self, app):
        try:
            if sys.version_info < (3, 10):  # pragma: no-cover-if-gte-py310
                # This reproduces the Python 3.10 platform.freedesktop_os_release() function.
                with self.tools.ETC_OS_RELEASE.open(encoding="utf-8") as f:
                    freedesktop_info = parse_freedesktop_os_release(f.read())
            else:  # pragma: no-cover-if-lt-py310
                freedesktop_info = self.tools.platform.freedesktop_os_release()

        except OSError as e:
            raise BriefcaseCommandError(
                "Could not find the /etc/os-release file. "
                "Is this a FreeDesktop-compliant Linux distribution?"
            ) from e

        return freedesktop_info

    def finalize_app_config(self, app: AppConfig):
        """Finalize app configuration.

        Linux .deb app configurations are deeper than other platforms, because they need
        to include components that are dependent on the target vendor and codename.
        Those properties are extracted from command-line options.

        The final app configuration merges the target-specific configuration into the
        generic "linux.deb" app configuration, as well as setting the Python version.

        :param app: The app configuration to finalize.
        """
        self.logger.info("Finalizing application configuration...", prefix=app.app_name)
        freedesktop_info = self.platform_freedesktop_info(app)

        # Process the FreeDesktop content to give the vendor, codename and vendor base.
        (
            app.target_vendor,
            app.target_codename,
            app.target_vendor_base,
        ) = self.vendor_details(freedesktop_info)

        self.logger.info(
            f"Targeting {app.target_vendor}:{app.target_codename} (Vendor base {app.target_vendor_base})"
        )

        if not self.use_docker:
            app.target_image = f"{app.target_vendor}:{app.target_codename}"
        else:
            # If we're building for Arch, and Docker does user mapping, we can't build,
            # because Arch won't let makepkg run as root. Docker on macOS *does* map the
            # user, but introducing a step-down user doesn't alter behavior, so we can
            # allow it.
            if (
                app.target_vendor_base == ARCH
                and self.tools.docker.is_user_mapped
                and self.tools.host_os != "Darwin"
            ):
                raise BriefcaseCommandError(
                    """\
Briefcase cannot use this Docker installation to target Arch Linux since the
tools to build packages for Arch cannot be run as root.

The Docker available to Briefcase requires the use of the root user in
containers to maintain accurate file permissions of the build artefacts.

This most likely means you're using Docker Desktop or rootless Docker.

Install Docker Engine and try again or run Briefcase on an Arch host system.
"""
                )

        # Merge target-specific configuration items into the app config This
        # means:
        # * merging app.linux.debian into app, overwriting anything global
        # * merging app.linux.ubuntu into app, overwriting anything vendor-base
        #   specific
        # * merging app.linux.ubuntu.focal into app, overwriting anything vendor
        #   specific
        # The vendor base config (e.g., redhat). The vendor base might not
        # be known, so fall back to an empty vendor config.
        if app.target_vendor_base:
            vendor_base_config = getattr(app, app.target_vendor_base, {})
        else:
            vendor_base_config = {}
        vendor_config = getattr(app, app.target_vendor, {})
        try:
            codename_config = vendor_config[app.target_codename]
        except KeyError:
            codename_config = {}

        # Copy all the specific configurations to the app config
        for config in [
            vendor_base_config,
            vendor_config,
            codename_config,
        ]:
            for key, value in config.items():
                setattr(app, key, value)

        with self.input.wait_bar("Determining glibc version..."):
            app.glibc_version = self.target_glibc_version(app)
        self.logger.info(f"Targeting glibc {app.glibc_version}")

        app.python_version_tag = self.app_python_version_tag(app)

        self.logger.info(f"Targeting Python{app.python_version_tag}")


class LinuxSystemMostlyPassiveMixin(LinuxSystemPassiveMixin):
    # The Mostly Passive mixin verifies that Docker exists and can be run, but
    # doesn't require that we're actually in a Linux environment.

    def app_python_version_tag(self, app: AppConfig):
        if self.use_docker:
            # If we're running in Docker, we can't know the Python3 version
            # before rolling out the template; so we fall back to "3". Later,
            # once we have a container in which we can run Python, this will be
            # updated to the actual Python version as part of the
            # `verify_python` app check.
            python_version_tag = "3"
        else:
            python_version_tag = super().app_python_version_tag(app)
        return python_version_tag

    def _build_env_abi(self, app: AppConfig):
        """Retrieves the ABI the packaging system is targeting in the build env.

        Each packaging system uses different values to identify the exact ABI that
        describes the target environment...so just defer to the packaging system.
        """
        command = {
            "deb": ["dpkg", "--print-architecture"],
            "rpm": ["rpm", "--eval", "%_target_cpu"],
            "pkg": ["pacman-conf", "Architecture"],
        }[app.packaging_format]
        try:
            return (
                self.tools[app].app_context.check_output(command).split("\n")[0].strip()
            )
        except (OSError, subprocess.CalledProcessError) as e:
            raise BriefcaseCommandError(
                "Failed to determine build environment's ABI for packaging."
            ) from e

    def deb_abi(self, app: AppConfig) -> str:
        """The default ABI for dpkg packaging for the target environment."""
        try:
            return self._deb_abi
        except AttributeError:
            self._deb_abi = self._build_env_abi(app)
            return self._deb_abi

    def rpm_abi(self, app: AppConfig) -> str:
        """The default ABI for rpm packaging for the target environment."""
        try:
            return self._rpm_abi
        except AttributeError:
            self._rpm_abi = self._build_env_abi(app)
            return self._rpm_abi

    def pkg_abi(self, app: AppConfig) -> str:
        """The default ABI for pacman packaging for the target environment."""
        try:
            return self._pkg_abi
        except AttributeError:
            self._pkg_abi = self._build_env_abi(app)
            return self._pkg_abi

    def distribution_filename(self, app: AppConfig) -> str:
        if app.packaging_format == "deb":
            return (
                f"{app.app_name}"
                f"_{app.version}"
                f"-{getattr(app, 'revision', 1)}"
                f"~{app.target_vendor}"
                f"-{app.target_codename}"
                f"_{self.deb_abi(app)}"
                ".deb"
            )
        elif app.packaging_format == "rpm":
            # openSUSE doesn't include a distro tag
            if app.target_vendor_base == SUSE:
                distro_tag = ""
            else:
                distro_tag = f".{self.rpm_tag(app)}"
            return (
                f"{app.app_name}"
                f"-{app.version}"
                f"-{getattr(app, 'revision', 1)}"
                f"{distro_tag}"
                f".{self.rpm_abi(app)}"
                f".rpm"
            )
        elif app.packaging_format == "pkg":
            return (
                f"{app.app_name}"
                f"-{app.version}"
                f"-{getattr(app, 'revision', 1)}"
                f"-{self.pkg_abi(app)}"
                ".pkg.tar.zst"
            )
        else:
            raise BriefcaseCommandError(
                "Briefcase doesn't currently know how to build system packages in "
                f"{app.packaging_format.upper()} format."
            )

    def distribution_path(self, app: AppConfig):
        return self.dist_path / self.distribution_filename(app)

    def target_glibc_version(self, app):
        """Determine the glibc version.

        If running in Docker, this is done by interrogating libc.so.6; outside docker,
        we can use os.confstr().
        """
        if self.use_docker:
            try:
                output = self.tools.docker.check_output(
                    ["ldd", "--version"],
                    image_tag=app.target_image,
                )
                # On Debian/Ubuntu, ldd --version will give you output of the form:
                #
                #     ldd (Ubuntu GLIBC 2.31-0ubuntu9.9) 2.31
                #     Copyright (C) 2020 Free Software Foundation, Inc.
                #     ...
                #
                # Other platforms produce output of the form:
                #
                #     ldd (GNU libc) 2.36
                #     Copyright (C) 2020 Free Software Foundation, Inc.
                #     ...
                #
                # Note that the exact text will vary version to version.
                # Look for the "2.NN" pattern.
                if match := re.search(r"\d\.\d+", output):
                    target_glibc = match.group(0)
                else:
                    raise BriefcaseCommandError(
                        "Unable to parse glibc dependency version from version string."
                    )
            except subprocess.CalledProcessError:
                raise BriefcaseCommandError(
                    "Unable to determine glibc dependency version."
                )

        else:
            target_glibc = super().target_glibc_version(app)

        return target_glibc

    def platform_freedesktop_info(self, app: AppConfig):
        if self.use_docker:
            # Preserve the target image on the command line as the app's target
            app.target_image = self.target_image

            # Extract release information from the image.
            with self.input.wait_bar(
                f"Checking Docker target image {app.target_image}..."
            ):
                output = self.tools.docker.check_output(
                    ["cat", "/etc/os-release"],
                    image_tag=app.target_image,
                )
                freedesktop_info = parse_freedesktop_os_release(output)
        else:
            freedesktop_info = super().platform_freedesktop_info(app)

        return freedesktop_info

    def docker_image_tag(self, app: AppConfig):
        """The Docker image tag for an app."""
        return f"briefcase/{app.bundle_identifier.lower()}:{app.target_vendor}-{app.target_codename}"

    def verify_tools(self):
        """If we're using Docker, verify that it is available."""
        super().verify_tools()
        if self.use_docker:
            Docker.verify(tools=self.tools, image_tag=self.target_image)

    def clone_options(self, command):
        """Clone the target_image option."""
        super().clone_options(command)
        self.target_image = command.target_image
        self.extra_docker_build_args = command.extra_docker_build_args

    def verify_python(self, app: AppConfig):
        """Verify that the version of Python being used to build the app in Docker is
        compatible with the version being used to run Briefcase.

        Will raise an exception if the Python version is fundamentally
        incompatible (i.e., if Briefcase doesn't support it); any other version
        discrepancy will log a warning, but continue.

        Requires that the app tools have been verified.

        As a side effect of verifying Python, the `python_version_tag` will be
        updated to reflect the *actual* python version, not just a generic "3".

        :param app: The application being built
        """
        output = self.tools[app].app_context.check_output(
            [
                f"python{app.python_version_tag}",
                "-c",
                (
                    "import sys; "
                    "print(f'{sys.version_info.major}.{sys.version_info.minor}')"
                ),
            ]
        )
        # Update the python version tag with the *actual* python version.
        app.python_version_tag = output.split("\n")[0]
        target_python_version = tuple(int(v) for v in app.python_version_tag.split("."))

        if target_python_version < self.briefcase_required_python_version:
            briefcase_min_version = ".".join(
                str(v) for v in self.briefcase_required_python_version
            )
            raise BriefcaseCommandError(
                f"The system python3 version provided by {app.target_image} "
                f"is {app.python_version_tag}; Briefcase requires a "
                f"minimum Python3 version of {briefcase_min_version}."
            )
        elif target_python_version != (
            self.tools.sys.version_info.major,
            self.tools.sys.version_info.minor,
        ):
            self.logger.warning(
                f"""
*************************************************************************
** WARNING: Python version mismatch!                                   **
*************************************************************************

    The system python3 provided by {app.target_image} is {app.python_version_tag}.
    This is not the same as your local system ({self.python_version_tag}).

    Ensure you have tested for Python version compatibility before
    releasing this app.

*************************************************************************
"""
            )

    def verify_system_python(self):
        """Verify that the Python being used to run Briefcase is the default system
        python.

        Will raise an exception if the system Python isn't an obvious Python3, or the
        Briefcase Python isn't the same version as the system Python.

        Requires that the app tools have been verified.
        """
        system_python_bin = Path("/usr/bin/python3").resolve()
        system_version = system_python_bin.name.split(".")
        if system_version[0] != "python3" or len(system_version) == 1:
            raise BriefcaseCommandError("Can't determine the system python version")

        if system_version[1] != str(self.tools.sys.version_info.minor):
            raise BriefcaseCommandError(
                f"The version of Python being used to run Briefcase ({self.python_version_tag}) "
                f"is not the system python3 (3.{system_version[1]})."
            )

    def _system_requirement_tools(self, app: AppConfig):
        """Utility method returning the packages and tools needed to verify system
        requirements.

        :param app: The app being built.
        :returns: A triple containing (0) The list of package names that must
            be installed at a bare minimum; (1) the arguments for the command
            used to verify the existence of a package on a system, and (2)
            the command used to install packages. All three values are `None`
            if the system cannot be identified.
        """
        if app.target_vendor_base == DEBIAN:
            base_system_packages = ["python3-dev", "build-essential"]
            system_verify = ["dpkg", "-s"]
            system_installer = ["apt", "install"]
        elif app.target_vendor_base == RHEL:
            base_system_packages = [
                "python3-devel",
                "gcc",
                "make",
                "pkgconf-pkg-config",
            ]
            system_verify = ["rpm", "-q"]
            system_installer = ["dnf", "install"]
        elif app.target_vendor_base == SUSE:
            base_system_packages = [
                "python3-devel",
                "patterns-devel-base-devel_basis",
            ]
            system_verify = ["rpm", "-q", "--whatprovides"]
            system_installer = ["zypper", "install"]
        elif app.target_vendor_base == ARCH:
            base_system_packages = [
                "python3",
                "base-devel",
            ]
            system_verify = ["pacman", "-Q"]
            system_installer = ["pacman", "-Syu"]
        else:
            base_system_packages = None
            system_verify = None
            system_installer = None

        return (
            base_system_packages,
            system_verify,
            system_installer,
        )

    def verify_system_packages(self, app: AppConfig):
        """Verify that the required system packages are installed.

        :param app: The app being built.
        """
        (
            base_system_packages,
            system_verify,
            system_installer,
        ) = self._system_requirement_tools(app)

        if not (system_verify and self.tools.shutil.which(system_verify[0])):
            self.logger.warning(
                """
*************************************************************************
** WARNING: Can't verify system packages                               **
*************************************************************************

    Briefcase doesn't know how to verify the installation of system
    packages on your Linux distribution. If you have any problems
    building this app, ensure that the packages listed in the app's
    `system_requires` setting have been installed.

*************************************************************************
"""
            )
            return

        # Run a check for each package listed in the app's system_requires,
        # plus the baseline system packages that are required.
        missing = []
        for package in base_system_packages + getattr(app, "system_requires", []):
            try:
                self.tools.subprocess.check_output(system_verify + [package])
            except subprocess.CalledProcessError:
                missing.append(package)

        # If any required packages are missing, raise an error.
        if missing:
            raise BriefcaseCommandError(
                f"""\
Unable to build {app.app_name} due to missing system dependencies. Run:

    sudo {" ".join(system_installer)} {" ".join(missing)}

to install the missing dependencies, and re-run Briefcase.
"""
            )

    def verify_app_tools(self, app: AppConfig):
        """Verify App environment is prepared and available.

        When Docker is used, create or update a Docker image for the App. Without
        Docker, the host machine will be used as the App environment.

        :param app: The application being built
        """
        # Verifying the App context is idempotent; but we have some
        # additional logic that we only want to run the first time through.
        # Check (and store) the pre-verify app tool state.
        verify_python = not hasattr(self.tools[app], "app_context")

        if self.use_docker:
            DockerAppContext.verify(
                tools=self.tools,
                app=app,
                image_tag=self.docker_image_tag(app),
                dockerfile_path=self.bundle_path(app) / "Dockerfile",
                app_base_path=self.base_path,
                host_bundle_path=self.bundle_path(app),
                host_data_path=self.data_path,
                python_version=app.python_version_tag,
                extra_build_args=self.extra_docker_build_args,
            )

            # Check the system Python on the target system to see if it is
            # compatible with Briefcase.
            if verify_python:
                self.verify_python(app)
        else:
            NativeAppContext.verify(tools=self.tools, app=app)

            # Check the system Python on the target system to see if it is
            # compatible with Briefcase, and that the required system packages
            # are installed.
            if verify_python:
                self.verify_system_python()
                self.verify_system_packages(app)

        # Establish Docker as app context before letting super set subprocess
        super().verify_app_tools(app)


class LinuxSystemMixin(LinuxSystemMostlyPassiveMixin):
    def verify_host(self):
        """If we're *not* using Docker, verify that we're actually on Linux."""
        super().verify_host()
        if not self.use_docker:
            if self.tools.host_os != "Linux":
                raise UnsupportedHostError(self.supported_host_os_reason)


class LinuxSystemCreateCommand(LinuxSystemMixin, LocalRequirementsMixin, CreateCommand):
    description = "Create and populate a Linux system project."

    def output_format_template_context(self, app: AppConfig):
        context = super().output_format_template_context(app)

        # Linux system templates use the target codename, rather than
        # the format "system" as the leaf of the bundle path
        context["format"] = app.target_codename

        # The base template context includes the host Python version;
        # override that with an app-specific Python version, allowing
        # for the app to be built with the system Python.
        context["python_version"] = app.python_version_tag

        # Add the docker base image
        context["docker_base_image"] = app.target_image

        # Add the vendor base
        context["vendor_base"] = app.target_vendor_base

        # Use the non-root user if Docker is not mapping usernames. Also use a non-root
        # user if we're on macOS; user mapping doesn't alter Docker operation, but some
        # packaging tools (e.g., Arch's makepkg) don't like running as root. If we're
        # not using Docker, this will fall back to the template default, which should be
        # enabling the root user. This might cause problems later, but it's part of a
        # much bigger "does the project need to be updated in light of configuration
        # changes" problem.
        try:
            context["use_non_root_user"] = (
                self.tools.host_os == "Darwin" or not self.tools.docker.is_user_mapped
            )
        except AttributeError:
            pass  # ignore if not using Docker

        return context


class LinuxSystemUpdateCommand(LinuxSystemCreateCommand, UpdateCommand):
    description = "Update an existing Linux system project."


class LinuxSystemOpenCommand(LinuxSystemMostlyPassiveMixin, DockerOpenCommand):
    description = (
        "Open a shell in a Docker container for an existing Linux system project."
    )


class LinuxSystemBuildCommand(LinuxSystemMixin, BuildCommand):
    description = "Build a Linux system project."

    def build_app(self, app: AppConfig, **kwargs):
        """Build an application.

        :param app: The application to build
        """
        self.logger.info("Building application...", prefix=app.app_name)

        self.logger.info("Build bootstrap binary...")
        with self.input.wait_bar("Building bootstrap binary..."):
            try:
                # Build the bootstrap binary.
                self.tools[app].app_context.run(
                    [
                        "make",
                        "-C",
                        "bootstrap",
                        "install",
                    ],
                    check=True,
                    cwd=self.bundle_path(app),
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Error building bootstrap binary for {app.app_name}."
                ) from e

        # Make the folder for docs
        doc_folder = (
            self.bundle_path(app)
            / f"{app.app_name}-{app.version}"
            / "usr"
            / "share"
            / "doc"
            / app.app_name
        )
        doc_folder.mkdir(parents=True, exist_ok=True)

        with self.input.wait_bar("Installing license..."):
            if license_file := app.license.get("file"):
                license_file = self.base_path / license_file
                if license_file.is_file():
                    self.tools.shutil.copy(license_file, doc_folder / "copyright")
                else:
                    raise BriefcaseCommandError(
                        f"""\
Your `pyproject.toml` specifies a license file of {str(license_file.relative_to(self.base_path))!r}.
However, this file does not exist.

Ensure you have correctly spelled the filename in your `license.file` setting.

"""
                    )
            elif license_text := app.license.get("text"):
                (doc_folder / "copyright").write_text(license_text, encoding="utf-8")
                if len(license_text.splitlines()) <= 1:
                    self.logger.warning(
                        """
Your app specifies a license using `license.text`, but the value doesn't appear to be a
full license. Briefcase will generate a `copyright` file for your project; you should
ensure that the contents of this file is adequate.
"""
                    )
            else:
                raise BriefcaseCommandError(
                    """\
Your project does not contain a LICENSE definition.

Create a file named `LICENSE` in the same directory as your `pyproject.toml`
with your app's licensing terms, and set `license.file = 'LICENSE'` in your
app's configuration.
"""
                )

        with self.input.wait_bar("Installing changelog..."):
            changelog = self.base_path / "CHANGELOG"
            if changelog.is_file():
                with changelog.open(encoding="utf-8") as infile:
                    outfile = gzip.GzipFile(
                        doc_folder / "changelog.gz", mode="wb", mtime=0
                    )
                    outfile.write(infile.read().encode("utf-8"))
                    outfile.close()
            else:
                raise BriefcaseCommandError(
                    """\
Your project does not contain a CHANGELOG file.

Create a file named `CHANGELOG` in the same directory as your `pyproject.toml`
with details about the release.
"""
                )

        # Make a folder for manpages
        man_folder = (
            self.bundle_path(app)
            / f"{app.app_name}-{app.version}"
            / "usr"
            / "share"
            / "man"
            / "man1"
        )
        man_folder.mkdir(parents=True, exist_ok=True)

        with self.input.wait_bar("Installing man page..."):
            manpage_source = self.bundle_path(app) / f"{app.app_name}.1"
            if manpage_source.is_file():
                with manpage_source.open(encoding="utf-8") as infile:
                    outfile = gzip.GzipFile(
                        man_folder / f"{app.app_name}.1.gz", mode="wb", mtime=0
                    )
                    outfile.write(infile.read().encode("utf-8"))
                    outfile.close()
            else:
                raise BriefcaseCommandError(
                    f"Template does not provide a manpage source file `{app.app_name}.1`"
                )

        self.logger.verbose("Update file permissions...")
        with self.input.wait_bar("Updating file permissions..."):
            for path in self.project_path(app).glob("**/*"):
                old_perms = self.tools.os.stat(path).st_mode & 0o777
                user_perms = old_perms & 0o700
                world_perms = old_perms & 0o007

                # File permissions like 775 and 664 (where the group and user
                # permissions are the same), cause Debian heartburn. So, make
                # sure the group and world permissions are the same
                new_perms = user_perms | (world_perms << 3) | world_perms

                # If there's been any change in permissions, apply them
                if new_perms != old_perms:  # pragma: no-cover-if-is-windows
                    self.logger.verbose(
                        "Updating file permissions on "
                        f"{path.relative_to(self.bundle_path(app))} "
                        f"from {oct(old_perms)[2:]} to {oct(new_perms)[2:]}"
                    )
                    path.chmod(new_perms)

        with self.input.wait_bar("Stripping binary..."):
            self.tools.subprocess.check_output(["strip", self.binary_path(app)])


class LinuxSystemRunCommand(LinuxSystemMixin, RunCommand):
    description = "Run a Linux system project."
    supported_host_os = {"Linux"}
    supported_host_os_reason = "Linux system projects can only be executed on Linux."

    def run_app(
        self,
        app: AppConfig,
        test_mode: bool,
        passthrough: list[str],
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param test_mode: Boolean; Is the app running in test mode?
        :param passthrough: The list of arguments to pass to the app
        """
        # Set up the log stream
        kwargs = self._prepare_app_kwargs(app=app, test_mode=test_mode)

        with self.tools[app].app_context.run_app_context(kwargs) as kwargs:
            # Console apps must operate in non-streaming mode so that console input can
            # be handled correctly. However, if we're in test mode, we *must* stream so
            # that we can see the test exit sentinel
            if app.console_app and not test_mode:
                self.logger.info("=" * 75)
                self.tools[app].app_context.run(
                    [self.binary_path(app)] + passthrough,
                    cwd=self.tools.home_path,
                    bufsize=1,
                    stream_output=False,
                    **kwargs,
                )
            else:
                # Start the app in a way that lets us stream the logs
                app_popen = self.tools[app].app_context.Popen(
                    [self.binary_path(app)] + passthrough,
                    cwd=self.tools.home_path,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    bufsize=1,
                    **kwargs,
                )

                # Start streaming logs for the app.
                self._stream_app_logs(
                    app,
                    popen=app_popen,
                    test_mode=test_mode,
                    clean_output=False,
                )


def debian_multiline_description(description):
    """Generate a Debian multiline description string.

    The long description in a Debian control file must *not* contain any blank lines,
    and each line must start with a single space. Convert a long description into Debian
    format.

    :param description: A multi-line long description string.
    :returns: A string in Debian's multiline format
    """
    return "\n ".join(line for line in description.split("\n") if line.strip() != "")


class LinuxSystemPackageCommand(LinuxSystemMixin, PackageCommand):
    description = "Package a Linux system project."

    @property
    def packaging_formats(self):
        return ["deb", "rpm", "pkg", "system"]

    def _verify_packaging_tools(self, app: AppConfig):
        """Verify that the local environment contains the packaging tools."""
        tool_name, executable_name, package_name = {
            "deb": ("dpkg", "dpkg-deb", "dpkg-dev"),
            "rpm": ("rpm-build", "rpmbuild", "rpm-build"),
            "pkg": ("makepkg", "makepkg", "pacman"),
        }[app.packaging_format]

        if not self.tools.shutil.which(executable_name):
            if install_cmd := self._system_requirement_tools(app)[2]:
                raise BriefcaseCommandError(
                    f"Can't find the {tool_name} tools. "
                    f"Try running `sudo {' '.join(install_cmd)} {package_name}`."
                )
            else:
                raise BriefcaseCommandError(
                    f"Can't find the {executable_name} tool. "
                    f"Install this first to package the {app.packaging_format}."
                )

    def verify_app_tools(self, app):
        super().verify_app_tools(app)
        # If "system" packaging format was selected, determine what that means.
        if app.packaging_format == "system":
            app.packaging_format = {
                DEBIAN: "deb",
                RHEL: "rpm",
                ARCH: "pkg",
                SUSE: "rpm",
            }.get(app.target_vendor_base, None)

        if app.packaging_format is None:
            raise BriefcaseCommandError(
                "Briefcase doesn't know the system packaging format for "
                f"{app.target_vendor}. You may be able to build a package "
                "by manually specifying a format with -p/--packaging-format"
            )

        if not self.use_docker:
            self._verify_packaging_tools(app)

    def package_app(self, app: AppConfig, **kwargs):
        if app.packaging_format == "deb":
            self._package_deb(app, **kwargs)
        elif app.packaging_format == "rpm":
            self._package_rpm(app, **kwargs)
        elif app.packaging_format == "pkg":
            self._package_pkg(app, **kwargs)
        else:
            raise BriefcaseCommandError(
                "Briefcase doesn't currently know how to build system packages in "
                f"{app.packaging_format.upper()} format."
            )

    def _package_deb(self, app: AppConfig, **kwargs):
        self.logger.info("Building .deb package...", prefix=app.app_name)

        # The long description *must* exist.
        if app.long_description is None:
            raise BriefcaseCommandError(
                "App configuration does not define `long_description`. "
                "Debian projects require a long description."
            )

        # Write the Debian metadata control file.
        with self.input.wait_bar("Write Debian package control file..."):
            DEBIAN_path = self.project_path(app) / "DEBIAN"

            if DEBIAN_path.exists():
                self.tools.shutil.rmtree(DEBIAN_path)

            DEBIAN_path.mkdir()
            # dpkg-dep requires "DEBIAN" directory is >=0755 and <=0775
            DEBIAN_path.chmod(0o755)

            # Add runtime package dependencies. App config has been finalized,
            # so this will be the target-specific definition, if one exists.
            # libc6 is added because lintian complains without it, even though
            # it's a dependency of the thing we *do* care about - python.
            system_runtime_requires = ", ".join(
                [
                    f"libc6 (>={app.glibc_version})",
                    f"libpython{app.python_version_tag}",
                ]
                + getattr(app, "system_runtime_requires", [])
            )

            with (DEBIAN_path / "control").open("w", encoding="utf-8") as f:
                f.write(
                    "\n".join(
                        [
                            f"Package: {app.app_name}",
                            f"Version: {app.version}",
                            f"Architecture: {self.deb_abi(app)}",
                            f"Maintainer: {app.author} <{app.author_email}>",
                            f"Homepage: {app.url}",
                            f"Description: {app.description}",
                            f" {debian_multiline_description(app.long_description)}",
                            f"Depends: {system_runtime_requires}",
                            f"Section: {getattr(app, 'system_section', 'utils')}",
                            "Priority: optional\n",
                        ]
                    )
                )

        with self.input.wait_bar("Building Debian package..."):
            try:
                # Build the dpkg.
                self.tools[app].app_context.run(
                    [
                        "dpkg-deb",
                        "--build",
                        "--root-owner-group",
                        f"{app.app_name}-{app.version}",
                    ],
                    check=True,
                    cwd=self.bundle_path(app),
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Error while building .deb package for {app.app_name}."
                ) from e

            # Move the deb file to its final location
            self.tools.shutil.move(
                self.bundle_path(app) / f"{app.app_name}-{app.version}.deb",
                self.distribution_path(app),
            )

    def _package_rpm(self, app: AppConfig, **kwargs):  # pragma: no-cover-if-is-windows
        self.logger.info("Building .rpm package...", prefix=app.app_name)

        # The long description *must* exist.
        if app.long_description is None:
            raise BriefcaseCommandError(
                "App configuration does not define `long_description`. "
                "Red Hat projects require a long description."
            )

        # Generate the rpmbuild layout
        rpmbuild_path = self.bundle_path(app) / "rpmbuild"
        with self.input.wait_bar("Generating rpmbuild layout..."):
            if rpmbuild_path.exists():
                self.tools.shutil.rmtree(rpmbuild_path)

            (rpmbuild_path / "BUILD").mkdir(parents=True)
            (rpmbuild_path / "BUILDROOT").mkdir(parents=True)
            (rpmbuild_path / "RPMS").mkdir(parents=True)
            (rpmbuild_path / "SOURCES").mkdir(parents=True)
            (rpmbuild_path / "SRPMS").mkdir(parents=True)
            (rpmbuild_path / "SPECS").mkdir(parents=True)

        # Add runtime package dependencies. App config has been finalized,
        # so this will be the target-specific definition, if one exists.
        system_runtime_requires = [
            "python3",
        ] + getattr(app, "system_runtime_requires", [])

        # Write the spec file
        with self.input.wait_bar("Write RPM spec file..."):
            with (rpmbuild_path / "SPECS" / f"{app.app_name}.spec").open(
                "w", encoding="utf-8"
            ) as f:
                f.write(
                    "\n".join(
                        [
                            # By default, rpmbuild thinks all .py files are executable,
                            # and if a .py doesn't have a shebang line, it will
                            # tell you that it will remove the executable bit -
                            # even if the executable bit isn't set.
                            # We disable the processor that does this.
                            "%global __brp_mangle_shebangs %{nil}",
                            # rpmbuild tries to strip binaries, which messes with
                            # binary wheels. Disable these checks.
                            "%global __brp_strip %{nil}",
                            "%global __brp_strip_static_archive %{nil}",
                            "%global __brp_strip_comment_note %{nil}",
                            # Disable RPATH checking, because check-rpaths can't deal with
                            # the structure of manylinux wheels
                            "%global __brp_check_rpaths %{nil}",
                            # Disable all the auto-detection that tries to magically
                            # determine requirements from the binaries
                            f"%global __requires_exclude_from ^%{{_libdir}}/{app.app_name}/.*$",
                            f"%global __provides_exclude_from ^%{{_libdir}}/{app.app_name}/.*$",
                            # Disable debug processing.
                            "%global _enable_debug_package 0",
                            "%global debug_package %{nil}",
                            "",
                            # Base package metadata
                            f"Name:           {app.app_name}",
                            f"Version:        {app.version}",
                            f"Release:        {getattr(app, 'revision', 1)}%{{?dist}}",
                            f"Summary:        {app.description}",
                            "",
                            "License:        Unknown",  # TODO: Add license information (see #1829)
                            f"URL:            {app.url}",
                            "Source0:        %{name}-%{version}.tar.gz",
                            "",
                        ]
                        + [
                            f"Requires:       {requirement}"
                            for requirement in system_runtime_requires
                        ]
                        + [
                            "",
                            f"ExclusiveArch:  {self.rpm_abi(app)}",
                            "",
                            "%description",
                            app.long_description,
                            "",
                            "%prep",
                            "%autosetup",
                            "",
                            "%build",
                            "",
                            "%install",
                            "cp -r usr %{buildroot}/usr",
                        ]
                    )
                )

                f.write("\n\n%files\n")
                # Build the file manifest. Include any file that is found; also include
                # any directory that includes an app_name component, as those paths
                # will need to be cleaned up afterwards. Files that *aren't*
                # in <app_name> (sub)directories (e.g., /usr/bin/<app_name> or
                # /usr/share/man/man1/<app_name>.1.gz) will be included, but paths
                # *not* cleaned up, as they're part of more general system structures.
                for filename in sorted(self.project_path(app).glob("**/*")):
                    path = filename.relative_to(self.project_path(app))

                    if filename.is_dir():
                        if app.app_name in path.parts:
                            f.write(f'%dir "/{path}"\n')
                    else:
                        f.write(f'"/{path}"\n')

                # Add the changelog content to the bottom of the spec file.
                f.write("\n%changelog\n")
                changelog_source = self.base_path / "CHANGELOG"
                if not changelog_source.is_file():
                    raise BriefcaseCommandError(
                        """\
Your project does not contain a CHANGELOG file.

Create a file named `CHANGELOG` in the same directory as your `pyproject.toml`
with details about the release.
"""
                    )
                with changelog_source.open(encoding="utf-8") as c:
                    f.write(c.read())

        with self.input.wait_bar("Building source archive..."):
            self.tools.shutil.make_archive(
                rpmbuild_path / "SOURCES" / f"{app.app_name}-{app.version}",
                format="gztar",
                root_dir=self.bundle_path(app),
                base_dir=f"{app.app_name}-{app.version}",
            )

        with self.input.wait_bar("Building RPM package..."):
            try:
                # Build the dpkg.
                self.tools[app].app_context.run(
                    [
                        "rpmbuild",
                        "-bb",
                        "--define",
                        f"_topdir {self.bundle_path(app) / 'rpmbuild'}",
                        f"./rpmbuild/SPECS/{app.app_name}.spec",
                    ],
                    check=True,
                    cwd=self.bundle_path(app),
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Error while building .rpm package for {app.app_name}."
                ) from e

        # Move the rpm file to its final location
        self.tools.shutil.move(
            rpmbuild_path
            / "RPMS"
            / self.rpm_abi(app)
            / self.distribution_filename(app),
            self.distribution_path(app),
        )

    def _package_pkg(self, app: AppConfig, **kwargs):  # pragma: no-cover-if-is-windows
        self.logger.info("Building .pkg.tar.zst package...", prefix=app.app_name)

        # The description *must* exist.
        # pkgdesc has 80 char limit.
        if app.description is None:
            raise BriefcaseCommandError(
                "App configuration does not define `description`. "
                "Arch projects require a description."
            )
        # The changelog should exist.
        changelog_source = self.base_path / "CHANGELOG"
        if not changelog_source.is_file():
            raise BriefcaseCommandError(
                """\
Your project does not contain a CHANGELOG file.
Create a file named `CHANGELOG` in the same directory as your `pyproject.toml`
with details about the release.
"""
            )

        # Generate the pkgbuild layout
        pkgbuild_path = self.bundle_path(app) / "pkgbuild"
        with self.input.wait_bar("Generating pkgbuild layout..."):
            if pkgbuild_path.exists():
                self.tools.shutil.rmtree(pkgbuild_path)
            pkgbuild_path.mkdir(parents=True)

            # Copy the CHANGELOG file to the pkgbuild_path so that it is visible to PKGBUILD
            self.tools.shutil.copy(changelog_source, pkgbuild_path / "CHANGELOG")

        # Build the source archive
        with self.input.wait_bar("Building source archive..."):
            self.tools.shutil.make_archive(
                pkgbuild_path / f"{app.app_name}-{app.version}",
                format="gztar",
                root_dir=self.bundle_path(app),
                base_dir=f"{app.app_name}-{app.version}",
            )

        # Write the arch PKGBUILD file.
        with self.input.wait_bar("Write PKGBUILD file..."):
            # Add runtime package dependencies. App config has been finalized,
            # so this will be the target-specific definition, if one exists.
            system_runtime_requires_list = [
                f"glibc>={app.glibc_version}",
                "python3",
            ] + getattr(app, "system_runtime_requires", [])

            system_runtime_requires = " ".join(
                f"'{pkg}'" for pkg in system_runtime_requires_list
            )

            with (pkgbuild_path / "PKGBUILD").open("w", encoding="utf-8") as f:
                f.write(
                    "\n".join(
                        [
                            f"# Maintainer: {app.author} <{app.author_email}>",
                            f'export PACKAGER="{app.author} <{app.author_email}>"',
                            f"pkgname={app.app_name}",
                            f"pkgver={app.version}",
                            f"pkgrel={getattr(app, 'revision', 1)}",
                            f'pkgdesc="{app.description}"',
                            f"arch=('{self.pkg_abi(app)}')",
                            f'url="{app.url}"',
                            "license=('Unknown')",
                            f"depends=({system_runtime_requires})",
                            "changelog=CHANGELOG",
                            'source=("$pkgname-$pkgver.tar.gz")',
                            "md5sums=('SKIP')",
                            "options=('!strip')",
                            "package() {",
                            '    cp -r "$srcdir/$pkgname-$pkgver/usr/" "$pkgdir"/usr/',
                            "}",
                        ]
                    )
                )

        with self.input.wait_bar("Building Arch package..."):
            try:
                # Build the pkg.
                self.tools[app].app_context.run(
                    [
                        "makepkg",
                    ],
                    env={"PKGEXT": ".pkg.tar.zst"},
                    check=True,
                    cwd=pkgbuild_path,
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Error while building .pkg.tar.zst package for {app.app_name}."
                ) from e

            # Move the pkg file to its final location
            self.tools.shutil.move(
                pkgbuild_path / self.distribution_filename(app),
                self.distribution_path(app),
            )


class LinuxSystemPublishCommand(LinuxSystemMixin, PublishCommand):
    description = "Publish a Linux system project."


# Declare the briefcase command bindings
create = LinuxSystemCreateCommand
update = LinuxSystemUpdateCommand
open = LinuxSystemOpenCommand
build = LinuxSystemBuildCommand
run = LinuxSystemRunCommand
package = LinuxSystemPackageCommand
publish = LinuxSystemPublishCommand



from unittest.mock import MagicMock

import pytest

import briefcase.commands.new


@pytest.fixture
def mock_select_option(monkeypatch):
    mock = MagicMock()
    monkeypatch.setattr(briefcase.commands.new, "_select_option", mock)
    return mock


def test_override_used(new_command, capsys):
    """The override is used if valid."""
    override_value = "value"
    assert (
        new_command.select_option(
            intro="intro",
            variable="variable",
            default=None,
            options=["value", "some_value", "other_value"],
            override_value="value",
        )
        == "value"
    )
    assert f"Using override value {override_value!r}" in capsys.readouterr().out


def test_override_validation(new_command, mock_select_option, capsys):
    """The override is not used if it is not a valid option."""
    new_command.select_option(
        intro="intro",
        variable="variable",
        default="value",
        options=["value", "some_value", "other_value"],
        override_value="invalid_value",
    )
    assert (
        "Invalid override value 'invalid_value' for Variable, using user-provided value."
        in capsys.readouterr().out
    )
    mock_select_option.assert_called_once_with(
        prompt="Variable [1]:",
        input=new_command.input,
        default="1",
        options=[
            ("value", "value"),
            ("some_value", "some_value"),
            ("other_value", "other_value"),
        ],
    )


def test_default_value_has_correct_index(new_command, mock_select_option):
    """The default value is used and has the correct index if it is a valid option."""
    new_command.select_option(
        intro="intro",
        variable="variable",
        default="some_value",
        options=["value", "some_value", "other_value"],
        override_value=None,
    )
    mock_select_option.assert_called_once_with(
        prompt="Variable [2]:",
        input=new_command.input,
        default="2",
        options=[
            ("value", "value"),
            ("some_value", "some_value"),
            ("other_value", "other_value"),
        ],
    )


def test_default_default_is_one(new_command, mock_select_option):
    """The default value is 1 if it is not provided."""
    new_command.select_option(
        intro="intro",
        variable="variable",
        default=None,
        options=["value", "some_value", "other_value"],
        override_value=None,
    )
    mock_select_option.assert_called_once_with(
        prompt="Variable [1]:",
        input=new_command.input,
        default="1",
        options=[
            ("value", "value"),
            ("some_value", "some_value"),
            ("other_value", "other_value"),
        ],
    )


def test_exception_if_wrong_default(new_command, mock_select_option):
    """An exception is raised if the default value is not a valid option."""
    with pytest.raises(ValueError):
        new_command.select_option(
            intro="intro",
            variable="variable",
            default="invalid_value",
            options=["value", "some_value", "other_value"],
            override_value=None,
        )



import pytest

from briefcase.exceptions import MissingToolError, NetworkFailure, UnsupportedHostError
from briefcase.integrations.base import ManagedTool
from briefcase.integrations.linuxdeploy import LinuxDeployBase
from tests.integrations.linuxdeploy.utils import (
    side_effect_create_mock_appimage,
    side_effect_create_mock_tool,
)


class LinuxDeployDummy(LinuxDeployBase, ManagedTool):
    name = "dummy-plugin"
    full_name = "Dummy plugin"
    install_msg = "Installing dummy plugin"

    def __init__(self, tools, file_name="linuxdeploy-dummy-wonky.AppImage", **kwargs):
        super().__init__(tools=tools)
        self._file_name = file_name
        self.kwargs = kwargs

    @property
    def file_name(self):
        return self._file_name

    @property
    def download_url(self):
        return f"https://example.com/path/to/{self._file_name}"

    @property
    def file_path(self):
        return self.tools.base_path / "somewhere"


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.linuxdeploy = "tool"

    tool = LinuxDeployDummy.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.linuxdeploy


@pytest.mark.parametrize("host_os", ["Windows", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{LinuxDeployDummy.name} is not supported on {host_os}",
    ):
        LinuxDeployDummy.verify(mock_tools)


def test_verify_exists(mock_tools, tmp_path):
    """If the tool/plugin already exists, verification doesn't download."""
    appimage_path = tmp_path / "tools/somewhere/linuxdeploy-dummy-wonky.AppImage"

    # Mock the existence of an install
    appimage_path.parent.mkdir(parents=True)
    appimage_path.touch()

    # Create a linuxdeploy wrapper by verification
    linuxdeploy = LinuxDeployDummy.verify(mock_tools)

    # No download occurred
    assert mock_tools.file.download.call_count == 0
    assert mock_tools.os.chmod.call_count == 0

    # The build command retains the path to the downloaded file.
    assert linuxdeploy.file_path == appimage_path.parent


def test_verify_does_not_exist_dont_install(mock_tools):
    """If the tool/plugin doesn't exist, and install=False, it is *not* downloaded."""
    # True to create a linuxdeploy wrapper by verification.
    # This will fail because it doesn't exist, but installation was disabled.
    with pytest.raises(MissingToolError):
        LinuxDeployDummy.verify(mock_tools, install=False)

    # No download occurred
    assert mock_tools.file.download.call_count == 0
    assert mock_tools.os.chmod.call_count == 0


def test_verify_does_not_exist(mock_tools, tmp_path):
    """If the tool/plugin doesn't exist, it is downloaded."""
    appimage_path = tmp_path / "tools/somewhere/linuxdeploy-dummy-wonky.AppImage"

    # Mock a successful download
    mock_tools.file.download.side_effect = side_effect_create_mock_appimage(
        appimage_path
    )

    # Create a linuxdeploy wrapper by verification, with some extra args
    linuxdeploy = LinuxDeployDummy.verify(mock_tools, arg1="value1", arg2="value2")

    # The extra verification arguments were passed to the tool instance
    assert linuxdeploy.kwargs == {
        "app": None,
        "arg1": "value1",
        "arg2": "value2",
    }

    # A download is invoked
    mock_tools.file.download.assert_called_with(
        url="https://example.com/path/to/linuxdeploy-dummy-wonky.AppImage",
        download_path=tmp_path / "tools/somewhere",
        role="Dummy plugin",
    )
    # The downloaded file will be made executable
    mock_tools.os.chmod.assert_called_with(appimage_path, 0o755)

    # The build command retains the path to the downloaded file.
    assert linuxdeploy.file_path == appimage_path.parent


def test_verify_does_not_exist_non_appimage(mock_tools, tmp_path):
    """If a non-Appimage tool/plugin doesn't exist, it is downloaded, but not elf-
    patched."""
    tool_path = tmp_path / "tools/somewhere/linuxdeploy-dummy.sh"

    # Mock a successful download
    mock_tools.file.download.side_effect = side_effect_create_mock_tool(tool_path)

    # Create a linuxdeploy wrapper by verification
    linuxdeploy = LinuxDeployDummy.verify(mock_tools, file_name=tool_path.name)

    # A download is invoked
    mock_tools.file.download.assert_called_with(
        url="https://example.com/path/to/linuxdeploy-dummy.sh",
        download_path=tmp_path / "tools/somewhere",
        role="Dummy plugin",
    )
    # The downloaded file will be made executable
    mock_tools.os.chmod.assert_called_with(tool_path, 0o755)

    # The tool content hasn't been altered
    with tool_path.open("r", encoding="utf-8") as f:
        assert f.read() == "I am a complete tool"

    # The build command retains the path to the downloaded file.
    assert linuxdeploy.file_path == tool_path.parent


def test_verify_linuxdeploy_download_failure(mock_tools, tmp_path):
    """If a tool/plugin doesn't exist, and a download failure occurs, an error is
    raised."""
    mock_tools.file.download.side_effect = NetworkFailure("mock")

    with pytest.raises(NetworkFailure, match="Unable to mock"):
        LinuxDeployDummy.verify(mock_tools)

    # A download was invoked
    mock_tools.file.download.assert_called_with(
        url="https://example.com/path/to/linuxdeploy-dummy-wonky.AppImage",
        download_path=tmp_path / "tools/somewhere",
        role="Dummy plugin",
    )



import inspect
import pkgutil
import sys
from typing import Dict, Set, Type

import pytest

import briefcase.integrations
from briefcase.integrations.base import ManagedTool, Tool, tool_registry


def integrations_modules() -> Set[str]:
    """All modules in ``briefcase.integrations`` irrespective of whether they are
    defined in ``briefcase.integrations.__all__``"""
    return {
        module.name
        for module in pkgutil.iter_modules(briefcase.integrations.__path__)
        if module.name not in {"base"}
    }


def tools_for_module(tool_module_name: str) -> Dict[str, Type[Tool]]:
    """Return classes that subclass Tool in a module in ``briefcase.integrations``, e.g.
    {"android_sdk": AndroidSDK}."""
    return dict(
        inspect.getmembers(
            sys.modules[f"briefcase.integrations.{tool_module_name}"],
            lambda klass: (
                inspect.isclass(klass)
                and issubclass(klass, (Tool, ManagedTool))
                and klass not in {Tool, ManagedTool}
            ),
        )
    )


@pytest.fixture
def all_defined_tools() -> Set[Type[Tool]]:
    """All classes under ``briefcase.integrations`` that subclass Tool."""
    return {
        tool
        for toolset in map(tools_for_module, integrations_modules())
        for tool in toolset.values()
    }


def test_tool_registry(all_defined_tools, simple_tools):
    """The Tool Registry must contain all defined Tools."""
    # test uses subset since registry will contain dummy testing tools
    assert all_defined_tools.issubset(tool_registry.values())


def test_unique_tool_names(all_defined_tools):
    """All tools must have a unique name."""
    assert len(all_defined_tools) == len({t.name for t in all_defined_tools})


def test_valid_tool_names(all_defined_tools):
    """All tools must have a name without spaces."""
    assert all(" " not in t.name for t in all_defined_tools)



import pytest

from briefcase.integrations.subprocess import ParseError, json_parser


@pytest.mark.parametrize(
    "data, output",
    [
        ('{"key": "value"}', {"key": "value"}),
        (b'{"key": "value"}', {"key": "value"}),
    ],
)
def test_json_parser_success(data, output):
    assert json_parser(data) == output


@pytest.mark.parametrize(
    "data",
    (
        'This is a prologue in my JSON output :( \n\n{"key": "value"}',
        b'This is a prologue in my JSON output :( \n\n{"key": "value"}',
    ),
)
def test_json_parser_fail(data):
    with pytest.raises(ParseError, match="Failed to parse output as JSON:"):
        json_parser(data)



import shutil
import subprocess
import sys
from pathlib import Path
from unittest.mock import MagicMock, call

import pytest

from briefcase.commands import CreateCommand
from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.docker import Docker, DockerAppContext
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux import LocalRequirementsMixin

from ...utils import create_file, create_tgz_file, create_zip_file


class DummyCreateCommand(LocalRequirementsMixin, CreateCommand):
    """A command that provides the stubs required to satisfy LocalRequirementsMixin."""

    platform = "Tester"
    output_format = "Dummy"

    def binary_path(self, app):
        return self.bundle_path(app) / f"{app.app_name}.bin"


@pytest.fixture
def no_docker_create_command(first_app_config, tmp_path):
    command = DummyCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )

    # Disable Docker use
    command.use_docker = False

    # Set the host architecture to something known for test purposes.
    command.tools.host_arch = "wonky"

    # Set the host system to Linux for test purposes
    command.tools.host_os = "Linux"

    # Mock the existence of Docker
    command.tools.subprocess = MagicMock(spec_set=Subprocess)

    # Mock shutil.copy to do the copy, but be observable
    command.tools.shutil.copy = MagicMock(side_effect=shutil.copy)

    command._briefcase_toml[first_app_config] = {
        "paths": {"app_packages_path": "path/to/app_packages"}
    }

    # At the time app requirements are installed, the project folder will exist.
    command.bundle_path(first_app_config).mkdir(parents=True, exist_ok=True)

    return command


@pytest.fixture
def create_command(no_docker_create_command, first_app_config, tmp_path, monkeypatch):
    # Enable Docker use
    no_docker_create_command.use_docker = True

    # Provide Docker
    monkeypatch.setattr(
        Docker, "_is_user_mapping_enabled", MagicMock(return_value=True)
    )
    no_docker_create_command.tools.docker = Docker(tools=no_docker_create_command.tools)
    # Provide Docker app context
    no_docker_create_command.tools[first_app_config].app_context = DockerAppContext(
        tools=no_docker_create_command.tools,
        app=first_app_config,
    )
    no_docker_create_command.tools[first_app_config].app_context.prepare(
        image_tag="briefcase/com.example.first-app:py3.X",
        dockerfile_path=tmp_path / "base_path/build/first-app/tester/dummy/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/tester/dummy",
        host_data_path=tmp_path / "briefcase",
        python_version="3.X",
    )

    # Reset the subprocess.run mock, removing the Docker setup call
    no_docker_create_command.tools.subprocess.run.reset_mock()

    return no_docker_create_command


@pytest.fixture
def first_package(tmp_path):
    # Create a local package to be built
    create_file(
        tmp_path / "local/first/setup.py",
        content="Python config",
    )
    create_file(
        tmp_path / "local/first/first.py",
        content="Python source",
    )

    return str(tmp_path / "local/first")


@pytest.fixture
def second_package(tmp_path):
    # Create a local pre-built sdist
    create_tgz_file(
        tmp_path / "local/second-2.3.4.tar.gz",
        content=[
            ("setup.py", "Python config"),
            ("second.py", "Python source"),
        ],
    )

    return str(tmp_path / "local/second-2.3.4.tar.gz")


@pytest.fixture
def third_package(tmp_path):
    # Create a local pre-built wheel
    create_zip_file(
        tmp_path / "local/third-3.4.5-py3-none-any.whl",
        content=[
            ("MANIFEST.in", "Wheel config"),
            ("third.py", "Python source"),
        ],
    )

    return str(tmp_path / "local/third-3.4.5-py3-none-any.whl")


@pytest.fixture
def other_package(create_command, first_app_config):
    # A stale sdist, built in a previous pass
    create_tgz_file(
        create_command.local_requirements_path(first_app_config)
        / "other_package-0.1.2.tar.gz",
        content=[
            ("setup.py", "Python config"),
            ("other.py", "Python source"),
        ],
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_install_app_requirements_in_docker(create_command, first_app_config, tmp_path):
    """If Docker is in use, a docker context is used to invoke pip."""

    # Install requirements
    create_command.install_app_requirements(first_app_config, test_mode=False)

    # pip was invoked inside docker.
    create_command.tools.subprocess.run.assert_called_once_with(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'base_path/build/first-app/tester/dummy'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "briefcase/com.example.first-app:py3.X",
            "python3.X",
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            "--target=/app/path/to/app_packages",
            "foo==1.2.3",
            "bar>=4.5",
        ],
        check=True,
        encoding="UTF-8",
        env={"DOCKER_CLI_HINTS": "false"},
    )

    # The local requirements path exists, but is empty
    local_requirements_path = create_command.local_requirements_path(first_app_config)
    assert local_requirements_path.exists()
    assert len(list(local_requirements_path.iterdir())) == 0


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_install_app_requirements_no_docker(
    no_docker_create_command,
    first_app_config,
    tmp_path,
):
    """If docker is *not* in use, calls are made on raw subprocess."""
    # Verify the tools; this should operate in the non-docker context
    no_docker_create_command.verify_tools()
    no_docker_create_command.verify_app_tools(first_app_config)

    # Install requirements
    no_docker_create_command.install_app_requirements(first_app_config, test_mode=False)

    # Docker is not verified.
    assert not hasattr(no_docker_create_command.tools, "docker")

    # Subprocess is used for app_context
    assert isinstance(
        no_docker_create_command.tools[first_app_config].app_context, Subprocess
    )
    assert (
        no_docker_create_command.tools[first_app_config].app_context
        is no_docker_create_command.tools.subprocess
    )

    # pip was invoked natively
    no_docker_create_command.tools[
        first_app_config
    ].app_context.run.assert_called_once_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={tmp_path / 'base_path/build/first-app/tester/dummy/path/to/app_packages'}",
            "foo==1.2.3",
            "bar>=4.5",
        ],
        check=True,
        encoding="UTF-8",
    )

    # The local requirements path exists, but is empty
    local_requirements_path = no_docker_create_command.local_requirements_path(
        first_app_config
    )
    assert local_requirements_path.exists()
    assert len(list(local_requirements_path.iterdir())) == 0


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_install_app_requirements_with_locals(
    create_command,
    first_app_config,
    tmp_path,
    first_package,  # A local folder to be built
    second_package,  # A pre-built sdist
    third_package,  # A pre-built wheel
    other_package,  # A stale local requirement
):
    """If the app has local requirements, they are compiled into sdists for
    installation."""
    # Add local requirements
    first_app_config.requires.extend([first_package, second_package, third_package])

    # Mock the side effect of building an sdist
    def build_sdist(*args, **kwargs):
        # Extract the folder name; assume that's the name of the package
        name = Path(args[0][-1]).name
        create_tgz_file(
            create_command.local_requirements_path(first_app_config)
            / f"{name}-1.2.3.tar.gz",
            content=[
                ("setup.py", "Python config"),
                ("local.py", "Python source"),
            ],
        )

    create_command.tools.subprocess.check_output.side_effect = build_sdist

    # Install requirements
    create_command.install_app_requirements(first_app_config, test_mode=False)

    # An sdist was built for the local package
    create_command.tools.subprocess.check_output.assert_called_once_with(
        [
            sys.executable,
            "-X",
            "utf8",
            "-m",
            "build",
            "--sdist",
            "--outdir",
            tmp_path / "base_path/build/first-app/tester/dummy/_requirements",
            str(tmp_path / "local/first"),
        ],
        encoding="UTF-8",
    )

    # An attempt was made to copy the prebuilt packages
    create_command.tools.shutil.copy.mock_calls = [
        call(
            str(tmp_path / "local/second-2.3.4.tar.gz"),
            tmp_path / "base_path/build/first-app/tester/dummy/_requirements",
        ),
        call(
            str(tmp_path / "local/third-3.4.5-py3-none-any.whl"),
            tmp_path / "base_path/build/first-app/tester/dummy/_requirements",
        ),
    ]

    # pip was invoked inside docker.
    create_command.tools.subprocess.run.assert_called_once_with(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'base_path/build/first-app/tester/dummy'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "briefcase/com.example.first-app:py3.X",
            "python3.X",
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            "--target=/app/path/to/app_packages",
            "foo==1.2.3",
            "bar>=4.5",
            "/app/_requirements/first-1.2.3.tar.gz",
            "/app/_requirements/second-2.3.4.tar.gz",
            "/app/_requirements/third-3.4.5-py3-none-any.whl",
        ],
        check=True,
        encoding="UTF-8",
        env={"DOCKER_CLI_HINTS": "false"},
    )

    # The local requirements path exists, and contains the compiled sdist, the
    # pre-existing sdist, and the pre-existing wheel; the old requirement has
    # been purged.
    local_requirements_path = create_command.local_requirements_path(first_app_config)
    assert local_requirements_path.exists()
    assert [f.name for f in sorted(local_requirements_path.iterdir())] == [
        "first-1.2.3.tar.gz",
        "second-2.3.4.tar.gz",
        "third-3.4.5-py3-none-any.whl",
    ]


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_install_app_requirements_with_bad_local(
    create_command,
    first_app_config,
    tmp_path,
    first_package,  # A local folder to be built
    other_package,  # A stale local requirement
):
    """If the app has local requirement that can't be built, an error is raised."""
    # Add a local requirement
    first_app_config.requires.append(first_package)

    # Mock the building an sdist raising an error
    create_command.tools.subprocess.check_output.side_effect = (
        subprocess.CalledProcessError(
            cmd=["python", "-m", "build", "..."], returncode=1
        )
    )

    # Install requirements
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to build sdist for .*/local/first",
    ):
        create_command.install_app_requirements(first_app_config, test_mode=False)

    # An attempt to build the sdist was made
    create_command.tools.subprocess.check_output.assert_called_once_with(
        [
            sys.executable,
            "-X",
            "utf8",
            "-m",
            "build",
            "--sdist",
            "--outdir",
            tmp_path / "base_path/build/first-app/tester/dummy/_requirements",
            str(tmp_path / "local/first"),
        ],
        encoding="UTF-8",
    )

    # pip was *not* invoked inside docker.
    create_command.tools.subprocess.run.assert_not_called()

    # The local requirements path exists, and is empty. It has been purged, but not refilled.
    local_requirements_path = create_command.local_requirements_path(first_app_config)
    assert local_requirements_path.exists()
    assert len(list(local_requirements_path.iterdir())) == 0


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_install_app_requirements_with_missing_local_build(
    create_command,
    first_app_config,
    tmp_path,
):
    """If the app references a requirement that needs to be built, but is missing, an
    error is raised."""
    # Define a local requirement, but don't create the files it points at
    first_app_config.requires.append(str(tmp_path / "local/first"))

    # Install requirements
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to find local requirement .*/local/first",
    ):
        create_command.install_app_requirements(first_app_config, test_mode=False)

    # No attempt to build the sdist was made
    create_command.tools.subprocess.check_output.assert_not_called()

    # pip was *not* invoked inside docker.
    create_command.tools.subprocess.run.assert_not_called()

    # The local requirements path exists, and is empty. It has been purged, but not refilled.
    local_requirements_path = create_command.local_requirements_path(first_app_config)
    assert local_requirements_path.exists()
    assert len(list(local_requirements_path.iterdir())) == 0


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_install_app_requirements_with_bad_local_file(
    create_command,
    first_app_config,
    tmp_path,
):
    """If the app references a local requirement file that doesn't exist, an error is
    raised."""
    # Add a local requirement that doesn't exist
    first_app_config.requires.append(str(tmp_path / "local/missing-2.3.4.tar.gz"))

    # Install requirements
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to find local requirement .*/local/missing-2.3.4.tar.gz",
    ):
        create_command.install_app_requirements(first_app_config, test_mode=False)

    # An attempt was made to copy the package
    create_command.tools.shutil.copy.assert_called_once_with(
        str(tmp_path / "local/missing-2.3.4.tar.gz"),
        tmp_path / "base_path/build/first-app/tester/dummy/_requirements",
    )

    # No attempt was made to build the sdist
    create_command.tools.subprocess.check_output.assert_not_called()

    # pip was *not* invoked inside docker.
    create_command.tools.subprocess.run.assert_not_called()

    # The local requirements path exists, and is empty. It has been purged, but not refilled.
    local_requirements_path = create_command.local_requirements_path(first_app_config)
    assert local_requirements_path.exists()
    assert len(list(local_requirements_path.iterdir())) == 0



from automation.bootstraps import BRIEFCASE_EXIT_SUCCESS_SIGNAL, EXIT_SUCCESS_NOTIFY
from briefcase.bootstraps import PySide6GuiBootstrap


class PySide6AutomationBootstrap(PySide6GuiBootstrap):
    def app_source(self):
        return f"""\
import importlib.metadata
import sys

from PySide6 import QtWidgets
from PySide6.QtCore import QTimer


class {{{{ cookiecutter.class_name }}}}(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("{{{{ cookiecutter.app_name }}}}")
        self.show()

        QTimer.singleShot(2000, self.exit_app)

    def exit_app(self):
        print("{EXIT_SUCCESS_NOTIFY}")
        print("{BRIEFCASE_EXIT_SUCCESS_SIGNAL}")
        QtWidgets.QApplication.quit()


def main():
    # Linux desktop environments use an app's .desktop file to integrate the app
    # in to their application menus. The .desktop file of this app will include
    # the StartupWMClass key, set to app's formal name. This helps associate the
    # app's windows to its menu item.
    #
    # For association to work, any windows of the app must have WMCLASS property
    # set to match the value set in app's desktop file. For PySide6, this is set
    # with setApplicationName().

    # Find the name of the module that was used to start the app
    app_module = sys.modules["__main__"].__package__
    # Retrieve the app's metadata
    metadata = importlib.metadata.metadata(app_module)

    QtWidgets.QApplication.setApplicationName(metadata["Formal-Name"])

    app = QtWidgets.QApplication(sys.argv)
    main_window = {{{{ cookiecutter.class_name }}}}()
    sys.exit(app.exec())
"""



import pytest


@pytest.mark.parametrize(
    "email",
    [
        "foo@example.com",
    ],
)
def test_valid_email(new_command, email):
    """Test that valid email addresses are accepted."""
    assert new_command.validate_email(email)


@pytest.mark.parametrize(
    "email",
    [
        "not a email address!",  # Free text.
    ],
)
def test_invalid_email(new_command, email):
    """Test that invalid email addresses are rejected."""
    with pytest.raises(ValueError):
        new_command.validate_email(email)



import os
from unittest.mock import MagicMock

import pytest

from briefcase.integrations.base import ToolCache
from briefcase.integrations.file import File
from briefcase.integrations.subprocess import Subprocess

WIX_DOWNLOAD_URL = "https://github.com/wixtoolset/wix3/releases/download/wix3141rtm/wix314-binaries.zip"


@pytest.fixture
def mock_tools(tmp_path, mock_tools) -> ToolCache:
    mock_tools.host_os = "Windows"
    mock_tools.os = MagicMock(spec=os)

    # Mock default tools
    mock_tools.subprocess = MagicMock(spec_set=Subprocess)
    mock_tools.file.download = MagicMock(spec_set=File.download)

    return mock_tools



import subprocess
import time
from unittest.mock import MagicMock

import pytest

from briefcase.integrations.subprocess import Subprocess

# hardcoded here since subprocess will only include these constants if Python is literally on Windows
CREATE_NO_WINDOW = 0x8000000
CREATE_NEW_PROCESS_GROUP = 0x200


@pytest.fixture
def streaming_process():
    """Mock process returned for Popen context manager in
    Subprocess._run_and_stream_output."""
    process = MagicMock()

    # Mock the readline values of an actual process. The final return value is "",
    # indicating that the process has exited; however, we insert a short sleep
    # to ensure that any other threads will have a chance to run before this
    # thread actually terminates.
    def mock_readline():
        yield from [
            "output line 1\n",
            "\n",
            "output line 3\n",
        ]
        time.sleep(0.1)
        yield ""

    process.stdout.readline.side_effect = mock_readline()
    process.poll.return_value = -3
    return process


@pytest.fixture
def mock_sub(mock_tools, streaming_process):
    mock_tools.os.environ = {
        "VAR1": "Value 1",
        "PS1": "\nLine 2\n\nLine 4",
        "PWD": "/home/user/",
    }

    sub = Subprocess(mock_tools)
    sub._subprocess = MagicMock(spec=subprocess)

    run_result = MagicMock(spec=subprocess.CompletedProcess)
    run_result.returncode = 0
    sub._subprocess.run.return_value = run_result

    sub._subprocess.check_output.return_value = "some output line 1\nmore output line 2"

    sub._subprocess.CREATE_NO_WINDOW = CREATE_NO_WINDOW
    sub._subprocess.CREATE_NEW_PROCESS_GROUP = CREATE_NEW_PROCESS_GROUP

    # Return the streaming process for output streaming
    sub._subprocess.Popen.return_value.__enter__.return_value = streaming_process

    return sub
